from policyengine_us.model_api import *
from policyengine_core.taxscales import MarginalRateTaxScale


class ny_supplemental_tax(Variable):
    value_type = float
    entity = TaxUnit
    label = "NY supplemental income tax"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.NY

    def formula(tax_unit, period, parameters):
        ny_taxable_income = tax_unit("ny_taxable_income", period)
        ny_agi = tax_unit("ny_agi", period)
        ny_main_income_tax = tax_unit("ny_main_income_tax", period)
        tax = parameters(period).gov.states.ny.tax.income
        p = tax.supplemental

        filing_status = tax_unit("filing_status", period)
        status = filing_status.possible_values
        statuses = [
            status.SINGLE,
            status.JOINT,
            status.HEAD_OF_HOUSEHOLD,
            status.SURVIVING_SPOUSE,
            status.SEPARATE,
        ]
        in_each_status = [filing_status == s for s in statuses]

        rates = tax.main
        single = rates.single
        joint = rates.joint
        hoh = rates.head_of_household
        surviving_spouse = rates.surviving_spouse
        separate = rates.separate
        scales = [single, joint, hoh, surviving_spouse, separate]

        previous_agi_threshold = select(
            in_each_status,
            [
                get_previous_threshold(ny_taxable_income, scale.thresholds)
                for scale in scales
            ],
        )

        applicable_amount = max_(
            0, ny_agi - max_(previous_agi_threshold, p.min_agi)
        )

        phase_in_fraction = min_(
            1,
            applicable_amount / p.phase_in_length,
        )

        # edge case for high agi
        agi_limit = select(
            in_each_status,
            [
                single.thresholds[-1],
                joint.thresholds[-1],
                hoh.thresholds[-1],
                surviving_spouse.thresholds[-1],
                separate.thresholds[-1],
            ],
        )
        high_agi_rate = select(
            in_each_status,
            [scale.marginal_rates(agi_limit + 1) for scale in scales],
        )

        supplemental_tax_high_agi = (
            ny_taxable_income * high_agi_rate - ny_main_income_tax
        )

        if p.in_effect:
            recapture_base = select(
                in_each_status,
                [
                    p.recapture_base.single.calc(ny_taxable_income),
                    p.recapture_base.joint.calc(ny_taxable_income),
                    p.recapture_base.head_of_household.calc(ny_taxable_income),
                    p.recapture_base.surviving_spouse.calc(ny_taxable_income),
                    p.recapture_base.separate.calc(ny_taxable_income),
                ],
            )

            incremental_benefit = select(
                in_each_status,
                [
                    p.incremental_benefit.single.calc(ny_taxable_income),
                    p.incremental_benefit.joint.calc(ny_taxable_income),
                    p.incremental_benefit.head_of_household.calc(
                        ny_taxable_income
                    ),
                    p.incremental_benefit.surviving_spouse.calc(
                        ny_taxable_income
                    ),
                    p.incremental_benefit.separate.calc(ny_taxable_income),
                ],
            )

            supplemental_tax_general = (
                recapture_base + phase_in_fraction * incremental_benefit
            )

            return where(
                ny_agi > agi_limit,
                supplemental_tax_high_agi,
                supplemental_tax_general,
            )

        return where(
            ny_agi > agi_limit,
            supplemental_tax_high_agi,
            0,
        )



from policyengine_us.model_api import *


class is_eligible_for_american_opportunity_credit(Variable):
    value_type = bool
    entity = Person
    label = "Eligible for American Opportunity Credit"
    documentation = "Whether the person is eligible for the AOC in respect of qualified tuition expenses for this tax year. The expenses must be for one of the first four years of post-secondary education, and the person must not have claimed the AOC for any four previous tax years."
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/25A#b_2"



from policyengine_us.model_api import *


class ma_tafdc_age_limit(Variable):
    value_type = float
    entity = Person
    label = "Applicable age limit for the Massachusetts Temporary Assistance for Families with Dependent Children (TAFDC)"
    definition_period = YEAR
    reference = (
        "https://www.law.cornell.edu/regulations/massachusetts/106-CMR-703-200"
    )
    defined_for = StateCode.MA

    def formula(person, period, parameters):
        p = parameters(
            period
        ).gov.states.ma.dta.tcap.tafdc.eligibility.age_limit
        is_in_secondary_school = person("is_in_secondary_school", period)
        # College attendees are not considered eligible dependents
        dependent_age_limit = where(
            is_in_secondary_school, p.student_dependent, p.dependent
        )
        is_pregnant = person("is_pregnant", period)
        teen_parent_age_limit = p.teen_parent
        return where(is_pregnant, teen_parent_age_limit, dependent_age_limit)



from policyengine_us.model_api import *


class ks_refundable_eitc(Variable):
    value_type = float
    entity = TaxUnit
    label = "Kansas refundable EITC amount"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.ksrevenue.gov/pdf/ip21.pdf"
        "https://www.ksrevenue.gov/pdf/ip22.pdf"
    )
    defined_for = StateCode.KS

    def formula(tax_unit, period, parameters):
        total_eitc = tax_unit("ks_total_eitc", period)
        nonrefundable_eitc = tax_unit("ks_nonrefundable_eitc", period)
        return max_(0, total_eitc - nonrefundable_eitc)



from policyengine_us.model_api import *


class farm_rent_income(Variable):
    value_type = float
    entity = Person
    label = "farm rental income"
    unit = USD
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/1402"
    uprating = "calibration.gov.irs.soi.farm_rent_income"



from policyengine_us.model_api import *


class il_ccap_enrolled(Variable):
    value_type = bool
    entity = SPMUnit
    definition_period = MONTH
    label = "Whether the family is currently enrolled in Illinois Child Care Assistance Program (CCAP)"
    defined_for = StateCode.IL



from policyengine_us.model_api import *


def create_dc_kccatc_reform(parameters, period, bypass=False):
    class dc_kccatc(Variable):
        value_type = float
        entity = TaxUnit
        label = "DC keep child care affordable tax credit"
        unit = USD
        definition_period = YEAR
        reference = (
            "https://otr.cfo.dc.gov/sites/default/files/dc/sites/otr/publication/attachments/52926_D-40_12.21.21_Final_Rev011122.pdf#page=67"
            "https://otr.cfo.dc.gov/sites/default/files/dc/sites/otr/publication/attachments/2022_D-40_Booklet_Final_blk_01_23_23_Ordc.pdf#page=59"
        )
        defined_for = StateCode.DC

        def formula(tax_unit, period, parameters):
            reform_parameters = parameters(period).gov.contrib.dc_kccatc

            if reform_parameters.active:
                # Eligible expenses

                expenses = tax_unit("tax_unit_childcare_expenses", period)
                cdcc = parameters(period).gov.irs.credits.cdcc
                # Cap at number of children based on federal CDCC limit.
                count_eligible = min_(
                    cdcc.eligibility.max,
                    tax_unit("count_cdcc_eligible", period),
                )

                # Maximum covered expenses
                max_expenses_per_child = reform_parameters.expenses.max
                max_expenses = max_expenses_per_child * count_eligible
                eligible_capped_expenses = min_(expenses, max_expenses)

                # Percent covered

                amount_covered = (
                    eligible_capped_expenses * reform_parameters.expenses.rate
                )

                # Phase-out
                agi = tax_unit("adjusted_gross_income", period)
                filing_status = tax_unit("filing_status", period)
                threshold = reform_parameters.phase_out.threshold[
                    filing_status
                ]
                income_over_threshold = max_(agi - threshold, 0)
                phase_out_rate = reform_parameters.phase_out.rate
                phase_out_amount = income_over_threshold * phase_out_rate

                return max_(0, amount_covered - phase_out_amount)
            else:
                p = parameters(period).gov.states.dc.tax.income.credits
                # determine tax unit's income eligibility status
                taxinc = tax_unit("dc_taxable_income_joint", period)
                filing_status = tax_unit("filing_status", period)
                income_eligible = (
                    taxinc <= p.kccatc.income_limit[filing_status]
                )
                # determine count of KCCATC age eligible children
                person = tax_unit.members
                is_dependent = person("is_tax_unit_dependent", period)
                age = person("age", period)
                kccatc_age_eligible = is_dependent & (age <= p.kccatc.max_age)
                kccatc_eligible_count = tax_unit.sum(kccatc_age_eligible)
                # determine count of federal CDCC age eligible children
                cdcc = parameters(period).gov.irs.credits.cdcc.eligibility
                cdcc_age_eligible = is_dependent & (age < cdcc.child_age)
                cdcc_eligible_count = tax_unit.sum(cdcc_age_eligible)
                # calculate KCCATC amount
                max_kccatc = kccatc_eligible_count * p.kccatc.max_amount
                total_care_expenses = tax_unit(
                    "tax_unit_childcare_expenses", period
                )
                ratio = np.zeros_like(cdcc_eligible_count)
                mask = cdcc_eligible_count > 0
                ratio[mask] = (
                    kccatc_eligible_count[mask] / cdcc_eligible_count[mask]
                )
                kccatc_care_expenses = total_care_expenses * ratio
                kccatc = min_(kccatc_care_expenses, max_kccatc)
                # return calculated kccatc amount if income eligible
                return income_eligible * kccatc

    class reform(Reform):
        def apply(self):
            self.update_variable(dc_kccatc)

    if bypass or parameters(period).gov.contrib.dc_kccatc.active:
        return reform
    else:
        return None


dc_kccatc_reform = create_dc_kccatc_reform(None, None, True)



from policyengine_us.model_api import *


class ChildcareProviderTypeGroup(Enum):
    DCC_SACC = "Licenced/registered/permitted day care center; registered school-age child care"
    FDC_GFDC = (
        "Registered family day care homes; licensed group family day care"
    )
    LE_GC = "Legally exempt group child care programs"
    LE_STD = "Informal child care standard rate"
    LE_ENH = "Informal child care enhanced rate"


class childcare_provider_type_group(Variable):
    value_type = Enum
    possible_values = ChildcareProviderTypeGroup
    # DCC_SACC is most common among provider types
    default_value = ChildcareProviderTypeGroup.DCC_SACC
    entity = Person
    label = "Childcare provider type group"
    definition_period = YEAR



from policyengine_us.model_api import *


class hi_cdcc_income_floor_eligible(Variable):
    value_type = bool
    entity = Person
    label = "Hawaii income floor eligible"
    defined_for = StateCode.HI
    unit = USD
    definition_period = YEAR
    reference = (
        "https://files.hawaii.gov/tax/forms/2022/n11ins.pdf#page=28"
        "https://files.hawaii.gov/tax/forms/2022/n11ins.pdf#page=29"
        "https://files.hawaii.gov/tax/legal/hrs/hrs_235.pdf#page=41"
        "https://files.hawaii.gov/tax/forms/2022/schx_i.pdf#page=2"
    )

    def formula(person, period, parameters):
        return person("is_disabled", period) | person(
            "is_full_time_student", period
        )



from policyengine_us.model_api import *


class il_aabd_countable_income(Variable):
    value_type = float
    entity = Person
    label = (
        "Illinois Aid to the Aged, Blind or Disabled (AABD) countable income"
    )
    unit = USD
    definition_period = MONTH
    defined_for = "il_aabd_non_financial_eligible_person"

    adds = [
        "il_aabd_earned_income_after_exemption_person",
        "il_aabd_countable_unearned_income",
    ]



from policyengine_us.model_api import *


class tx_tanf_dependent_care_deduction(Variable):
    value_type = float
    entity = SPMUnit
    label = "Texas TANF dependent care deduction"
    unit = USD
    definition_period = MONTH
    reference = (
        "https://www.hhs.texas.gov/handbooks/texas-works-handbook/a-1420-types-deductions",
        "https://www.law.cornell.edu/regulations/texas/1-Tex-Admin-Code-SS-372-409",
    )
    defined_for = StateCode.TX

    def formula(spm_unit, period, parameters):
        # Actual cost of dependent child care, capped at maximum by age
        # Per § 372.409 (a)(3): up to $200/month for children under 2, $175/month for children 2+

        # Get actual childcare expenses
        childcare_expenses = spm_unit("childcare_expenses", period)

        # Calculate maximum deduction for dependents (children or incapacitated adults)
        person = spm_unit.members
        is_dependent = person("is_tax_unit_dependent", period)
        is_incapacitated = person("is_incapable_of_self_care", period)
        age = person("monthly_age", period)

        # Eligible for deduction: dependent child OR incapacitated adult
        eligible_for_deduction = is_dependent | is_incapacitated

        p = parameters(period).gov.states.tx.tanf.income
        max_per_person = where(
            eligible_for_deduction, p.deductions.dependent_care.calc(age), 0
        )
        total_max_deduction = spm_unit.sum(max_per_person)

        # Deduction is the lesser of actual expenses or maximum
        return min_(childcare_expenses, total_max_deduction)



from policyengine_us.model_api import *


class id_retirement_benefits_deduction_relevant_income(Variable):
    value_type = float
    entity = Person
    label = "Idaho retirement benefits deduction income sources"
    unit = USD
    reference = "https://legislature.idaho.gov/statutesrules/idstat/title63/t63ch30/sect63-3022a/"
    definition_period = YEAR
    defined_for = "id_retirement_benefits_deduction_eligible_person"

    adds = "gov.states.id.tax.income.deductions.retirement_benefits.income_sources"



from policyengine_us.model_api import *


class or_eitc(Variable):
    value_type = float
    entity = TaxUnit
    label = "OR EITC"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.oregon.gov/dor/forms/FormsPubs/form-or-40-inst_101-040-1_2021.pdf#page=18",
        "https://www.oregonlegislature.gov/bills_laws/ors/ors315.html",  # Subsection 315.266
    )
    defined_for = StateCode.OR

    def formula(tax_unit, period, parameters):
        # Grab federal EITC.
        federal_eitc = tax_unit("eitc", period)
        # Check if the tax unit has a young child to qualify for a higher rate.
        person = tax_unit.members
        p = parameters(period).gov.states["or"].tax.income.credits.eitc
        # Check if any person in the simulation in a young child.
        young_child = person("age", period) < p.old_child_age
        # Re-aggregate to the tax unit level.
        has_young_child = tax_unit.any(young_child)
        # Multiply by the relevant factor.
        rate = where(
            has_young_child, p.match.has_young_child, p.match.no_young_child
        )
        return federal_eitc * rate



from policyengine_us.model_api import *
from policyengine_core.periods import period as period_


def create_mi_surtax() -> Reform:
    class mi_surtax(Variable):
        value_type = float
        entity = TaxUnit
        label = "Michigan surtax"
        defined_for = StateCode.MI
        unit = USD
        definition_period = YEAR

        def formula(tax_unit, period, parameters):
            taxable_income = tax_unit("mi_taxable_income", period)
            joint = tax_unit("tax_unit_is_joint", period)
            p = parameters(period).gov.contrib.states.mi.surtax.rate
            return where(
                joint,
                p.joint.calc(taxable_income),
                p.single.calc(taxable_income),
            )

    class mi_income_tax(Variable):
        value_type = float
        entity = TaxUnit
        label = "Michigan income tax"
        unit = USD
        definition_period = YEAR
        defined_for = StateCode.MI

        adds = ["mi_income_tax_before_refundable_credits", "mi_surtax"]
        subtracts = ["mi_refundable_credits"]

    class reform(Reform):
        def apply(self):
            self.update_variable(mi_income_tax)
            self.update_variable(mi_surtax)

    return reform


def create_mi_surtax_reform(parameters, period, bypass: bool = False):
    if bypass:
        return create_mi_surtax()

    p = parameters.gov.contrib.states.mi.surtax

    reform_active = False
    current_period = period_(period)

    for i in range(5):
        if p(current_period).in_effect:
            reform_active = True
            break
        current_period = current_period.offset(1, "year")

    if reform_active:
        return create_mi_surtax()
    else:
        return None


mi_surtax = create_mi_surtax_reform(None, None, bypass=True)



from policyengine_us.model_api import *


class mn_wfc_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Minnesota working family credit eligibilty status"
    definition_period = YEAR
    reference = (
        "https://www.revisor.mn.gov/statutes/2021/cite/290.0671"
        "https://www.revisor.mn.gov/statutes/cite/290.0671"
    )
    defined_for = StateCode.MN

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.mn.tax.income.credits.cwfc
        person = tax_unit.members
        # determine demographic eligibility using WFC rules
        has_child = tax_unit("tax_unit_children", period) > 0
        age = person("age", period)
        min_age = p.wfc.eligible.childless_adult_age.minimum
        max_age = p.wfc.eligible.childless_adult_age.maximum
        in_age_range = (age >= min_age) & (age <= max_age)
        age_eligible = in_age_range & ~person("is_tax_unit_dependent", period)
        demographic_eligible = has_child | tax_unit.any(age_eligible)
        # determine investment income eligibility using federal EITC rules
        invinc_eligible = tax_unit("eitc_investment_income_eligible", period)
        # determine if tax unit has separate filing status
        filing_status = tax_unit("filing_status", period)
        separate = filing_status == filing_status.possible_values.SEPARATE
        # determine WFC eligibility
        return demographic_eligible & invinc_eligible & ~separate



from policyengine_us.model_api import *


class care_and_support_costs(Variable):
    value_type = float
    entity = Person
    unit = USD
    definition_period = YEAR
    label = "Total costs for this person's care and support"



from policyengine_core.model_api import *
from policyengine_us.entities import *
from policyengine_us.tools.branched_simulation import BranchedSimulation
from pathlib import Path
import pandas as pd
from policyengine_us.typing import Formula

USD = "currency-USD"


def tax_unit_non_dep_sum(var, tax_unit, period):
    return tax_unit.sum(
        tax_unit.members(var, period)
        * not_(tax_unit.members("is_tax_unit_dependent", period))
    )


def sum_contained_tax_units(var, population, period):
    tax_unit = population.members.tax_unit.reference_entity
    values = tax_unit(var, period)
    is_head = population.members("is_tax_unit_head", period)
    person_level_values = tax_unit.project(values) * is_head
    return population.sum(person_level_values)


infinity = np.inf
select = np.select
where = np.where

PERCENT = "/1"


def variable_alias(name: str, variable_cls: type) -> type:
    """
    Copy a variable class and return a new class.
    """
    class_dict = dict(variable_cls.__dict__)
    class_dict["formula"] = lambda entity, period: entity(
        variable_cls.__name__, period
    )
    return type(
        name,
        variable_cls.__bases__,
        class_dict,
    )


def sum_among_non_dependents(variable: str) -> Callable:
    def formula(tax_unit, period, parameters):
        return tax_unit_non_dep_sum(variable, tax_unit, period)

    return formula


def spouse(person: Population, period: int, variable: str) -> ArrayLike:
    values = person(variable, period)
    return (person.marital_unit.sum(values) - values).astype(values.dtype)


def in_state(state):
    def is_eligible(population, period, parameters):
        return population("state_code_str", period) == state

    return is_eligible


def get_next_threshold(values: ArrayLike, thresholds: ArrayLike) -> ArrayLike:
    """
    Return the next threshold in the sequence of thresholds.
    """
    t = np.array(thresholds)
    return t[
        min_((t <= values.reshape((1, len(values))).T).sum(axis=1), len(t) - 1)
    ]


def get_previous_threshold(
    values: ArrayLike, thresholds: ArrayLike
) -> ArrayLike:
    """
    Return the previous threshold in the sequence of thresholds.
    """
    t = np.array(thresholds)
    return t[
        max_((t <= values.reshape((1, len(values))).T).sum(axis=1) - 1, 0)
    ]



from policyengine_us.model_api import *


class sc_eitc(Variable):
    value_type = float
    entity = TaxUnit
    label = "South Carolina EITC"
    unit = USD
    definition_period = YEAR
    reference = "https://dor.sc.gov/forms-site/Forms/TC60_2021.pdf"
    defined_for = StateCode.SC

    def formula(tax_unit, period, parameters):
        federal_eitc = tax_unit("eitc", period)
        rate = parameters(period).gov.states.sc.tax.income.credits.eitc.rate
        return np.round(federal_eitc * rate, 1)



from policyengine_us.model_api import *


class self_employed_health_insurance_premiums(Variable):
    value_type = float
    entity = Person
    label = "Self-employed health insurance premiums"
    unit = USD
    documentation = "Health insurance premiums for plans covering individuals who are not covered by any employer-sponsored health insurance."
    definition_period = YEAR
    defined_for = "is_self_employed"
    adds = ["health_insurance_premiums"]



from policyengine_us.model_api import *


class ct_property_tax_credit_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Eligible for the Connecticut Property Tax Credit"
    definition_period = YEAR
    defined_for = StateCode.CT
    # (b)(2)
    reference = "https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-704c"

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.ct.tax.income.credits.property_tax
        age_eligible = (
            tax_unit("greater_age_head_spouse", period) >= p.age_threshold
        )
        p = parameters(period).gov.states.ct.tax.income.credits.property_tax
        dependents_present = tax_unit("tax_unit_dependents", period) > 0

        return dependents_present | age_eligible



from policyengine_us.model_api import *


class dc_deduction_joint(Variable):
    value_type = float
    entity = TaxUnit
    label = "DC deduction for each tax unit"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.DC

    def formula(tax_unit, period, parameters):
        # DC individual income tax FAQs include this:
        # Q: If I claimed itemized deductions on my federal return, must I
        #    also itemize on my DC return?
        # A: Yes. If you claim itemized deductions on your federal tax return,
        #    you must itemize on your DC tax return. You must take the same
        #    type of deduction (itemized or standard) on your DC return as
        #    taken on your federal return.
        # https://otr.cfo.dc.gov/page/individual-income-tax-special-circumstances-faqs
        return where(
            tax_unit("tax_unit_itemizes", period),
            tax_unit("dc_itemized_deductions", period),
            tax_unit("dc_standard_deduction", period),
        )



from policyengine_us.model_api import *


class ctc_phase_in_relevant_earnings(Variable):
    value_type = float
    entity = TaxUnit
    label = "Relevant earnings for the CTC phase-in"
    unit = USD
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/24#d"

    def formula(tax_unit, period, parameters):

        ctc = parameters(period).gov.irs.credits.ctc

        earnings = tax_unit("tax_unit_earned_income", period)
        earnings_over_threshold = max_(
            0, earnings - ctc.refundable.phase_in.threshold
        )
        return earnings_over_threshold * ctc.refundable.phase_in.rate



from policyengine_us.model_api import *


class CaCalworksChildCareFactorCategory(Enum):
    STANDARD = "Standard Rate Ceilings"
    EVENING_AND_WEEKEND_I = (
        "Evening/Weekend Care Rate Ceilings (50% or more of time)"
    )
    EVENING_AND_WEEKEND_II = "Evening/Weekend Care Rate Ceilings (at least 10% but less than 50% of time)"
    EXCEPTIONAL_NEEDS = "Exceptional Needs Care Rate Ceilings"
    SEVERELY_DISABLED = "Severely Disabled Care Rate Ceilings"


class ca_calworks_child_care_factor_category(Variable):
    value_type = Enum
    possible_values = CaCalworksChildCareFactorCategory
    default_value = CaCalworksChildCareFactorCategory.STANDARD
    entity = Person
    label = "California CalWORKs Child Care factor category"
    definition_period = MONTH
    defined_for = StateCode.CA
    reference = "http://epolicy.dpss.lacounty.gov/epolicy/epolicy/server/general/projects_responsive/ePolicyMaster/index.htm?&area=general&type=responsivehelp&ctxid=&project=ePolicyMaster#t=mergedProjects%2FChild%20Care%2FChild_Care%2F1210_8_Regional_Market_Rate_Ceilings%2F1210_8_Regional_Market_Rate_Ceilings.htm%23Referencesbc-11&rhtocid=_3_3_8_10"



from policyengine_us.model_api import *


class ne_school_readiness_credit(Variable):
    value_type = float
    entity = Person
    label = "Nebraska school readiness tax credit"
    unit = USD
    definition_period = YEAR
    reference = "https://revenue.nebraska.gov/sites/default/files/doc/tax-forms/2024/f_Individual_Income_Tax_Booklet.pdf#page=2"

    defined_for = "ne_school_readiness_credit_eligible_worker"

    def formula(person, period, parameters):
        level = person(
            "ne_school_readiness_credit_child_care_worker_rating", period
        )
        p = parameters(
            period
        ).gov.states.ne.tax.income.credits.school_readiness
        capped_level = min_(level, p.max_unit_size)
        # determine school readiness credit amount
        return p.amount.refundable[capped_level]



from policyengine_us.model_api import *


class vt_standard_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Vermont standard deduction"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://tax.vermont.gov/sites/tax/files/documents/IN-111-2022.pdf",  # Line4
        "http://legislature.vermont.gov/statutes/section/32/151/05811",  # Titl. 32 V.S.A. § 5811(21)(C)(ii)(iii)
    )
    defined_for = StateCode.VT

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.vt.tax.income.deductions.standard
        # Base deduction based on filing status.
        filing_status = tax_unit("filing_status", period)
        base_deduction = p.base[filing_status]
        # Vermont mirrors the federal definition of aged/blind by citing 26 U.S.C. § 63(f).
        # The aged_blind_count variable captures this, for head and spouse.
        aged_blind_count = tax_unit("aged_blind_count", period)
        aged_blind_deduction = aged_blind_count * p.additional
        return base_deduction + aged_blind_deduction



from policyengine_us.model_api import *


class nyc_refundable_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "NYC refundable credits"
    unit = USD
    definition_period = YEAR
    defined_for = "in_nyc"
    reference = "https://www.tax.ny.gov/pit/credits/new_york_city_credits.htm#:~:text=New%20for%202022,adjusted%20gross%20income%20(NYAGI)."

    adds = "gov.local.ny.nyc.tax.income.credits.refundable"



from policyengine_us.model_api import *


class sc_2022_rebate(Variable):
    value_type = float
    entity = TaxUnit
    label = "South Carolina 2022 rebate"
    defined_for = StateCode.SC
    unit = USD
    definition_period = YEAR
    reference = "https://www.scstatehouse.gov/sess124_2021-2022/bills/1087.htm"

    def formula(tax_unit, period, parameters):
        tax_before_non_refundable_credits = tax_unit(
            "sc_income_tax_before_non_refundable_credits", period
        )
        p = parameters(period).gov.states.sc.tax.income.credits
        other_non_refundable_credits = 0
        for credit in p.non_refundable:
            if credit != "sc_2022_rebate":
                other_non_refundable_credits += tax_unit(credit, period)
        refundable_credits = tax_unit("sc_refundable_credits", period)
        tax_after_credits = max_(
            0,
            tax_before_non_refundable_credits
            - other_non_refundable_credits
            - refundable_credits,
        )
        return min_(tax_after_credits, p["2022_rebate"].amount)



from policyengine_us.model_api import *


class itemized_deductions_less_salt(Variable):
    value_type = float
    entity = TaxUnit
    label = "Itemized tax deductions other than SALT deduction"
    unit = USD
    documentation = "Non-SALT itemized deductions total."
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.irs.deductions
        deductions = [
            deduction
            for deduction in p.itemized_deductions
            if deduction not in ["salt_deduction"]
        ]
        return add(tax_unit, period, deductions)



"""Test unified uprating extensions through 2100."""


def test_all_uprating_factors_extend_to_2100():
    """Test that all uprating factors extend through 2100 with consistent growth rates."""
    from policyengine_us import Microsimulation

    sim = Microsimulation()
    parameters = sim.tax_benefit_system.parameters

    # Define all uprating parameters to test with their specific periods
    uprating_params = [
        ("IRS", parameters.gov.irs.uprating, "-01-01"),
        ("SNAP", parameters.gov.usda.snap.uprating, "-10-01"),
        ("SSA", parameters.gov.ssa.uprating, "-01-01"),
        ("HHS", parameters.gov.hhs.uprating, "-01-01"),
        ("CPI-U", parameters.gov.bls.cpi.cpi_u, "-02-01"),
        ("Chained CPI-U", parameters.gov.bls.cpi.c_cpi_u, "-02-01"),
        ("CPI-W", parameters.gov.bls.cpi.cpi_w, "-02-01"),
    ]

    for name, param, date_suffix in uprating_params:
        # Test that values exist and are positive for future years
        test_years = [2035, 2050, 2075, 2100]
        values = []

        for year in test_years:
            value = param(f"{year}{date_suffix}")
            assert (
                value > 0
            ), f"No positive {name} uprating value for year {year}"
            values.append(value)

        # Test that values are monotonically increasing
        for i in range(1, len(values)):
            assert (
                values[i] > values[i - 1]
            ), f"{name} uprating should increase from {test_years[i-1]} to {test_years[i]}"

        # Test that growth is consistent in the extended period
        # Use years after the projection period ends
        year1, year2, year3 = 2040, 2041, 2042
        val1 = param(f"{year1}{date_suffix}")
        val2 = param(f"{year2}{date_suffix}")
        val3 = param(f"{year3}{date_suffix}")

        growth_rate_1 = val2 / val1
        growth_rate_2 = val3 / val2

        # Growth rates should be approximately equal (within 0.1%)
        assert (
            abs(growth_rate_1 - growth_rate_2) < 0.001
        ), f"{name} growth rate should be consistent: {growth_rate_1:.5f} vs {growth_rate_2:.5f}"


def test_uprating_growth_rates_are_reasonable():
    """Test that all uprating growth rates are within reasonable bounds."""
    from policyengine_us import Microsimulation

    sim = Microsimulation()
    parameters = sim.tax_benefit_system.parameters

    # Annual growth rates should be between 0.5% and 5% for inflation measures
    min_annual_growth = 1.005
    max_annual_growth = 1.05

    # Test all uprating parameters
    uprating_params = [
        ("IRS", parameters.gov.irs.uprating, "-01-01"),
        ("SNAP", parameters.gov.usda.snap.uprating, "-10-01"),
        ("SSA", parameters.gov.ssa.uprating, "-01-01"),
        ("HHS", parameters.gov.hhs.uprating, "-01-01"),
        ("CPI-U", parameters.gov.bls.cpi.cpi_u, "-02-01"),
        ("Chained CPI-U", parameters.gov.bls.cpi.c_cpi_u, "-02-01"),
        ("CPI-W", parameters.gov.bls.cpi.cpi_w, "-02-01"),
    ]

    year1, year2 = 2050, 2051

    for name, param, date_suffix in uprating_params:
        value_year1 = param(f"{year1}{date_suffix}")
        value_year2 = param(f"{year2}{date_suffix}")
        growth_rate = value_year2 / value_year1

        assert (
            min_annual_growth <= growth_rate <= max_annual_growth
        ), f"{name} growth rate {growth_rate:.4f} outside reasonable bounds [{min_annual_growth:.3f}, {max_annual_growth:.3f}]"


def test_cpi_relationships():
    """Test that CPI indices maintain expected relationships."""
    from policyengine_us import Microsimulation

    sim = Microsimulation()
    parameters = sim.tax_benefit_system.parameters

    # Test a few years to ensure relationships are maintained
    test_years = [2040, 2060, 2080, 2100]

    for year in test_years:
        cpi_u = parameters.gov.bls.cpi.cpi_u(f"{year}-02-01")
        c_cpi_u = parameters.gov.bls.cpi.c_cpi_u(f"{year}-02-01")

        # Chained CPI-U typically grows slower than regular CPI-U
        # due to substitution effects, but not always
        # Just verify both exist and are positive
        assert cpi_u > 0, f"CPI-U should be positive in {year}"
        assert c_cpi_u > 0, f"Chained CPI-U should be positive in {year}"



from policyengine_us.model_api import *


class hi_alternative_tax_on_capital_gains_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Eligible for the Hawaii alternative tax on capital gains"
    definition_period = YEAR
    defined_for = StateCode.HI

    # The tax form spcifies an income threshold which is not defined in the law
    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.hi.tax.income.alternative_tax
        net_capital_gains = tax_unit("net_capital_gain", period)
        return p.availability & (net_capital_gains > 0)



from policyengine_us.model_api import *


class amt_base_tax(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "Alternative Minimum Tax base tax"
    unit = USD
    documentation = "Alternative Minimum Tax (AMT) base tax, Form 6251 Part II Line 7 'All Others'"
    reference = "https://www.irs.gov/pub/irs-pdf/f6251.pdf"

    adds = [
        "amt_lower_base_tax",
        "amt_higher_base_tax",
    ]



from policyengine_us.model_api import *
from policyengine_us.variables.gov.hhs.tax_unit_fpg import fpg


class il_fpp_income_level(Variable):
    value_type = float
    entity = Person
    label = "Illinois Family Planning Program income level"
    unit = "/1"
    definition_period = YEAR
    reference = (
        "https://www.dhs.state.il.us/page.aspx?item=146077",
        "https://hfs.illinois.gov/medicalclients/familyplanning.html",
    )
    defined_for = StateCode.IL

    def formula(person, period, parameters):
        # Only the applicant's income is counted (not household income)
        income = person("adjusted_gross_income_person", period)

        # Fixed household size of 2 per IL FPP policy
        p = parameters(period).gov.states.il.hfs.fpp.eligibility

        # Calculate FPG using fixed household size
        state_group = person.household("state_group_str", period)
        fpp_fpg = fpg(p.household_size, state_group, period, parameters)

        # Return income as fraction of FPL
        return income / fpp_fpg



from policyengine_us.model_api import *
from numpy import ceil


class md_ctc(Variable):
    value_type = float
    entity = TaxUnit
    label = "Maryland Child Tax Credit"
    definition_period = YEAR
    unit = USD
    reference = [
        "https://casetext.com/statute/code-of-maryland/article-tax-general/title-10-income-tax/subtitle-7-income-tax-credits/section-10-751-effective-until-712026-tax-credit-for-qualified-child",
        "https://law.justia.com/codes/maryland/2022/tax-general/title-10/subtitle-7/section-10-751/",
        "https://mgaleg.maryland.gov/2025RS/Chapters_noln/CH_604_hb0352e.pdf#page=169",  # Maryland House Bill 352 - Budget Reconciliation and Financing Act of 2025
    ]
    defined_for = "md_ctc_eligible"

    def formula_2020(tax_unit, period, parameters):
        p = parameters(period).gov.states.md.tax.income.credits.ctc
        person = tax_unit.members
        dependent = person("is_tax_unit_dependent", period)
        disabled = person("is_disabled", period)
        age_limit = where(
            disabled, p.age_threshold.disabled, p.age_threshold.main
        )
        meets_age_limit = person("age", period) < age_limit
        eligible = dependent & meets_age_limit
        eligible_children = tax_unit.sum(eligible)

        # Calculate base credit amount
        base_credit = eligible_children * p.amount

        # Apply federal CTC reduction if applicable
        if p.reduced_by_federal_credit:
            federal_ctc = tax_unit("ctc", period)
            base_credit = max_(base_credit - federal_ctc, 0)

        # Apply income-based phase-out instead of hard cutoff when phase-out applies
        if p.phase_out.applies:
            agi = tax_unit("adjusted_gross_income", period)

            # Calculate phase-out reduction: reduction for every increment (or fraction thereof) above threshold
            excess_income = max_(agi - p.phase_out.threshold, 0)
            # Use ceiling division: any fraction of increment counts as a full increment for phase-out
            increments_above_threshold = ceil(
                excess_income / p.phase_out.increment
            )
            phase_out_amount = increments_above_threshold * p.phase_out.rate

            # Apply phase-out (credit cannot go below zero)
            return max_(base_credit - phase_out_amount, 0)

        return base_credit



from policyengine_us.model_api import *


class dc_self_employment_loss_addition(Variable):
    value_type = float
    entity = Person
    label = "DC excess self-employment loss addition"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://otr.cfo.dc.gov/sites/default/files/dc/sites/otr/publication/attachments/52926_D-40_12.21.21_Final_Rev011122.pdf#page=63"
        "https://otr.cfo.dc.gov/sites/default/files/dc/sites/otr/publication/attachments/2022_D-40_Booklet_Final_blk_01_23_23_Ordc.pdf#page=55"
    )
    defined_for = StateCode.DC

    def formula(person, period, parameters):
        loss_person = max_(0, -person("self_employment_income", period))
        loss_taxunit = person.tax_unit.sum(loss_person)
        p = parameters(period).gov.states.dc.tax.income.additions
        addition_taxunit = max_(
            0, loss_taxunit - p.self_employment_loss.threshold
        )
        # allocate taxunit addition in proportion to head and spouse losses
        filing_status = person.tax_unit("filing_status", period)
        is_joint = filing_status == filing_status.possible_values.JOINT
        loss_fraction = np.zeros_like(loss_person)
        mask = loss_taxunit > 0
        loss_fraction[mask] = loss_person[mask] / loss_taxunit[mask]
        addition_fraction = where(is_joint, loss_fraction, 1)
        is_head = person("is_tax_unit_head", period)
        is_spouse = person("is_tax_unit_spouse", period)
        return (is_head | is_spouse) * addition_taxunit * addition_fraction



from policyengine_us.model_api import *


class savers_credit_qualified_contributions(Variable):
    value_type = float
    entity = Person
    label = "Retirement Savings Credit qualified contributions"
    unit = USD
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/25B#d_2"

    def formula(person, period, parameters):
        p = parameters(period).gov.irs.credits.retirement_saving
        contributions = add(
            person, period, p.qualified_retirement_savings_contributions
        )
        distributions = add(person, period, ["retirement_distributions"])
        return max_(contributions - distributions, 0)



from policyengine_us.model_api import *


class il_property_tax_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "Illinois Property Tax Credit"
    unit = USD
    definition_period = YEAR

    defined_for = "il_is_exemption_eligible"

    def formula(tax_unit, period, parameters):
        ptax_paid = add(tax_unit, period, ["real_estate_taxes"])
        pre_credit_tax = tax_unit(
            "il_income_tax_before_non_refundable_credits", period
        )
        p = parameters(period).gov.states.il.tax.income.credits
        return min_(ptax_paid * p.property_tax.rate, pre_credit_tax)



from policyengine_us.model_api import *


class or_agi(Variable):
    value_type = float
    entity = TaxUnit
    label = "Oregon adjusted gross income"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.OR
    adds = ["adjusted_gross_income", "or_additions"]
    subtracts = ["or_income_subtractions"]



from policyengine_us.model_api import *


def create_dc_tax_threshold_joint_ratio_reform(
    parameters, period, bypass=False
):
    if not bypass:
        joint_ratio = parameters(
            period
        ).gov.contrib.dc_tax_threshold_joint_ratio
    else:
        joint_ratio = None

    class dc_income_tax_before_credits_joint(Variable):
        value_type = float
        entity = TaxUnit
        label = (
            "DC income tax before credits when married couples file jointly"
        )
        unit = USD
        definition_period = YEAR
        reference = (
            "https://otr.cfo.dc.gov/sites/default/files/dc/sites/otr/publication/attachments/52926_D-40_12.21.21_Final_Rev011122.pdf#page=36"
            "https://otr.cfo.dc.gov/sites/default/files/dc/sites/otr/publication/attachments/2022_D-40_Booklet_Final_blk_01_23_23_Ordc.pdf#page=34"
        )
        defined_for = StateCode.DC

        def formula(tax_unit, period, parameters):
            taxinc = max_(0, tax_unit("dc_taxable_income_joint", period))
            joint_ratio = parameters(
                period
            ).gov.contrib.dc_tax_threshold_joint_ratio
            p = parameters(period).gov.states
            return (
                p.dc.tax.income.rates.calc(taxinc / joint_ratio) * joint_ratio
            )

    class reform(Reform):
        def apply(self):
            self.update_variable(dc_income_tax_before_credits_joint)

    if bypass or joint_ratio != 1:
        return reform
    else:
        return None


dc_tax_threshold_joint_ratio_reform = (
    create_dc_tax_threshold_joint_ratio_reform(None, None, True)
)



from policyengine_us.model_api import *


class il_ccap_immigration_status_eligible_person(Variable):
    value_type = bool
    entity = Person
    label = "Eligible person for Illinois Child Care Assistance Program (CCAP) based on immigration status"
    definition_period = MONTH
    reference = "https://www.dhs.state.il.us/page.aspx?item=46885"
    defined_for = StateCode.IL

    def formula(person, period, parameters):
        p = parameters(period).gov.states.il.dhs.ccap
        immigration_status = person("immigration_status", period)
        immigration_status_str = immigration_status.decode_to_str()
        has_qualifying_status = np.isin(
            immigration_status_str,
            p.qualified_alien_statuses,
        )
        is_citizen = (
            immigration_status == immigration_status.possible_values.CITIZEN
        )

        return has_qualifying_status | is_citizen



from policyengine_us.model_api import *


class free_school_meals_reported(Variable):
    value_type = float
    entity = SPMUnit
    definition_period = YEAR
    label = "Free school meals (reported)"
    unit = USD



from policyengine_us.model_api import *


class local_sales_tax(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "Local sales tax"
    unit = USD

    def formula(tax_unit, period, parameters):
        # Until modeling the local sales tax table, estimate by multiplying by 0.2.
        return tax_unit("state_sales_tax", period) * 0.2



from policyengine_us.model_api import *


class wa_tanf_countable_earned_income(Variable):
    value_type = float
    entity = SPMUnit
    label = "Washington TANF countable earned income"
    unit = USD
    definition_period = MONTH
    reference = "https://app.leg.wa.gov/wac/default.aspx?cite=388-450-0170"
    defined_for = StateCode.WA

    def formula(spm_unit, period, parameters):
        # Get gross earned income from federal TANF variable
        gross_earned = add(spm_unit, period, ["tanf_gross_earned_income"])

        # Apply earned income disregard per WAC 388-450-0170:
        # "We start by deducting the first $500 of the total household's
        # earned income. We then subtract 50% of the remaining monthly
        # gross earned income."
        p = parameters(
            period
        ).gov.states.wa.dshs.tanf.income.deductions.earned_income_disregard

        # Step 1: Deduct flat $500 family earnings disregard
        remainder = max_(gross_earned - p.amount, 0)

        # Step 2: Subtract (disregard) 50% of remaining income
        # Per WAC 388-450-0170(3): "subtract 50% of the remaining...income"
        amount_disregarded = remainder * p.percentage_disregarded

        # Countable earned income = remainder - amount disregarded
        return remainder - amount_disregarded



from policyengine_us.model_api import *


class ky_taxable_income_joint(Variable):
    value_type = float
    entity = Person
    label = "Kentucky taxable income when married couples file jointly"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.KY
    reference = "https://revenue.ky.gov/Forms/740%20Packet%20Instructions%205-9-23.pdf#page=11"

    def formula(person, period, parameters):
        ky_agi = person("ky_agi", period)
        is_head = person("is_tax_unit_head", period)
        total_agi = is_head * person.tax_unit.sum(ky_agi)
        deductions = add(person.tax_unit, period, ["ky_deductions_joint"])

        return max_(0, total_agi - deductions)



from policyengine_us.model_api import *


class il_income_tax_rebate(Variable):
    value_type = float
    entity = TaxUnit
    label = "Illinois income tax rebate"
    defined_for = "il_income_tax_rebate_eligible"
    unit = USD
    definition_period = YEAR
    reference = "https://codes.findlaw.com/il/chapter-35-revenue/il-st-sect-35-5-212-1.html"

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.il.tax.income.credits.income_tax_rebate
        federal_agi = tax_unit("adjusted_gross_income", period)
        joint = tax_unit("tax_unit_is_joint", period)
        base_amount = where(
            joint,
            p.amount.joint.calc(federal_agi),
            p.amount.other.calc(federal_agi),
        )
        dependent_count = tax_unit("tax_unit_dependents", period)
        capped_dependents = min_(dependent_count, p.max_dependents)
        dependent_amount = capped_dependents * p.amount.dependent
        return base_amount + dependent_amount



from policyengine_us.model_api import *


class in_harris_county_tx(Variable):
    value_type = bool
    entity = Household
    definition_period = YEAR
    label = "Is in Harris County, Texas"

    def formula(household, period, parameters):
        county = household("county_str", period)
        return county == "HARRIS_COUNTY_TX"



from policyengine_us.model_api import *


class ca_calworks_child_care_welfare_to_work(Variable):
    value_type = float
    entity = Person
    label = "California CalWORKs Welfare to Work"
    unit = "hour"
    definition_period = YEAR
    defined_for = StateCode.CA
    reference = [
        "http://epolicy.dpss.lacounty.gov/epolicy/epolicy/server/general/projects_responsive/ePolicyMaster/index.htm?&area=general&type=responsivehelp&ctxid=&project=ePolicyMaster#t=mergedProjects%2FChild%20Care%2FChild_Care%2F1210_Overview%2F1210_Overview.htm%23Backgroundbc-3&rhtocid=_3_3_0_2",
        "https://leginfo.legislature.ca.gov/faces/codes_displaySection.xhtml?lawCode=WIC&sectionNum=11322.6.",
    ]



from policyengine_us.model_api import *


class poverty_line(Variable):
    value_type = float
    entity = SPMUnit
    label = "poverty line"
    unit = USD
    documentation = "Income threshold below which a household is considered to be in poverty."
    definition_period = YEAR
    adds = [
        "spm_unit_spm_threshold",
    ]



from policyengine_us.model_api import *


class TXCCSProviderRating(Enum):
    REG = "Regular"
    TRS2 = "Texas Rising Star 2"
    TRS3 = "Texas Rising Star 3"
    TRS4 = "Texas Rising Star 4"
    TSR = "Texas School Ready"
    NONE = "None"


class tx_ccs_provider_rating(Variable):
    value_type = Enum
    possible_values = TXCCSProviderRating
    default_value = TXCCSProviderRating.REG
    entity = Person
    definition_period = MONTH
    label = "Texas Child Care Services (CCS) provider rating"
    defined_for = StateCode.TX
    reference = "https://www.twc.texas.gov/sites/default/files/ccel/docs/bcy25-board-max-provider-payment-rates-4-age-groups-twc.pdf"



from policyengine_us.model_api import *


class ny_ctc_post_2024(Variable):
    value_type = float
    entity = TaxUnit
    label = "New York CTC post-2024"
    documentation = "New York's Empire State Child Credit under post-2024 rules (2025-2027)"
    unit = USD
    definition_period = YEAR
    reference = "https://www.nysenate.gov/legislation/laws/TAX/606"  # (c-1)
    defined_for = StateCode.NY

    def formula(tax_unit, period):
        base_credit = tax_unit("ny_ctc_post_2024_base", period)
        phase_out = tax_unit("ny_ctc_post_2024_phase_out", period)

        # Ensure credit cannot go negative
        return max_(base_credit - phase_out, 0)



from policyengine_us.model_api import *


class az_529_college_savings_plan_subtraction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Arizona 529 college savings plan subtraction"
    unit = USD
    documentation = "https://azdor.gov/sites/default/files/2023-03/FORMS_INDIVIDUAL_2022_140i.pdf#page=15"
    definition_period = YEAR
    defined_for = StateCode.AZ

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.az.tax.income.subtractions.college_savings

        filing_status = tax_unit("az_filing_status", period)
        contributions_529 = tax_unit("investment_in_529_plan", period)

        cap_per_beneficiary = p.cap[filing_status]
        beneficiaries = add(
            tax_unit, period, ["count_529_contribution_beneficiaries"]
        )
        cap = cap_per_beneficiary * beneficiaries

        return min_(contributions_529, cap)



from policyengine_us.model_api import *


class loss_limited_net_capital_gains(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "Loss-limited net capital gains"
    unit = USD

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.irs
        filing_status = tax_unit("filing_status", period)
        loss_limit = p.capital_gains.loss_limit[filing_status]
        net_capital_gains = tax_unit("net_capital_gains", period)
        return max_(-loss_limit, net_capital_gains)



from policyengine_us.model_api import *


class nm_property_tax_rebate(Variable):
    value_type = float
    entity = TaxUnit
    label = "New Mexico property tax rebate"
    unit = USD
    definition_period = YEAR
    reference = "https://nmonesource.com/nmos/nmsa/en/item/4340/index.do#!fragment/zoupio-_Toc140503750/BQCwhgziBcwMYgK4DsDWszIQewE4BUBTADwBdoAvbRABwEtsBaAfX2zgEYAWABgFYeAZgDsAgJQAaZNlKEIARUSFcAT2gBydRIiEwuBIuVrN23fpABlPKQBCagEoBRADKOAagEEAcgGFHE0jAAI2hSdjExIA"
    defined_for = StateCode.NM
    defined_for = "nm_property_tax_rebate_eligible"

    def formula(tax_unit, period, parameters):
        # Get property tax paid
        ptax_owner = add(tax_unit, period, ["real_estate_taxes"])
        # Get rent and multiply by 6%
        rent = add(tax_unit, period, ["rent"])
        p = parameters(period).gov.states.nm.tax.income.rebates.property_tax
        rent_percent = rent * p.rent_rate
        rent_and_ptax = ptax_owner + rent_percent
        # Get the maximum property tax liability
        agi = tax_unit("nm_modified_gross_income", period)
        max_liability = p.max_property_tax_liability.calc(agi, right=True)
        rebate = max_(0, rent_and_ptax - max_liability)
        # Maximum amount is based on filing status
        filing_status = tax_unit("filing_status", period)
        return min_(rebate, p.max_amount[filing_status])



from policyengine_us.model_api import *


class sc_net_capital_gain_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "South Carolina net capital gain deduction"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.SC
    reference = (
        "https://dor.sc.gov/forms-site/Forms/IITPacket_2021.pdf#page=15",
        "https://www.scstatehouse.gov/code/t12c006.php",
        # South Carolina Code of Laws Section 12-6-1150 (A)
    )

    def formula(tax_unit, period, parameters):
        ltcg = add(tax_unit, period, ["long_term_capital_gains"])
        stcg = add(tax_unit, period, ["short_term_capital_gains"])
        capped_stcg = min_(0, stcg)
        base = max_(0, ltcg + capped_stcg)
        p = parameters(
            period
        ).gov.states.sc.tax.income.deductions.net_capital_gain
        return base * p.rate



from policyengine_us.model_api import *


class rents(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Tax unit pays rent"
    unit = USD
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        return add(tax_unit, period, ["rent"]) > 0



from policyengine_us.model_api import *


class in_agi_tax(Variable):
    value_type = float
    entity = TaxUnit
    label = "Indiana adjusted gross income tax"
    definition_period = YEAR
    unit = USD
    reference = "http://iga.in.gov/legislative/laws/2021/ic/titles/006#6-3-2-1"  # (a)(3)
    defined_for = StateCode.IN

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states["in"].tax.income
        in_agi = tax_unit("in_agi", period)
        return max_(0, in_agi * p.agi_rate)



from policyengine_us.model_api import *


class self_employment_tax(Variable):
    value_type = float
    entity = Person
    label = "self-employment tax"
    definition_period = YEAR
    unit = USD

    def formula(person, period, parameters):
        if parameters(
            period
        ).gov.contrib.ubi_center.flat_tax.abolish_self_emp_tax:
            return 0
        return add(
            person,
            period,
            [
                "self_employment_social_security_tax",
                "self_employment_medicare_tax",
            ],
        )



from policyengine_us.model_api import *


class ca_eitc(Variable):
    value_type = float
    entity = TaxUnit
    label = "CalEITC"
    unit = USD
    definition_period = YEAR
    defined_for = "ca_eitc_eligible"

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.ca.tax.income.credits.earned_income

        earned_income = tax_unit("filer_adjusted_earnings", period)
        child_count = tax_unit("eitc_child_count", period)

        phase_in_rate = p.phase_in.rate.calc(child_count) * p.adjustment.factor
        phase_in_max_income = p.earned_income_amount.calc(child_count)

        phase_in_income = min_(earned_income, phase_in_max_income)
        phased_in_amount = phase_in_income * phase_in_rate

        phase_out_min_income = p.phase_out.start.calc(child_count)
        phase_out_rate = (
            p.phase_out.rate.calc(child_count) * p.adjustment.factor
        )

        phase_out_income = max_(0, earned_income - phase_out_min_income)

        second_phase_out_start_eitc = p.phase_out.final.start.calc(
            child_count
        )  # Expressed as the EITC amount at which the second phase-out starts,
        # not the income level.

        maximum_eitc = phase_in_max_income * phase_in_rate
        eitc_fall_by_first_phase_out = (
            maximum_eitc - second_phase_out_start_eitc
        )
        earnings_range_of_first_phase_out = (
            eitc_fall_by_first_phase_out / phase_out_rate
        )
        second_phase_out_start = (
            phase_out_min_income + earnings_range_of_first_phase_out
        )
        second_phase_out_end = p.phase_out.final.end

        phase_out_income = min_(
            phase_out_income, earnings_range_of_first_phase_out
        )
        amount_after_first_phase_out = (
            phased_in_amount - phase_out_income * phase_out_rate
        )
        percentage_along_second_phase_out = min_(
            (earned_income - second_phase_out_start)
            / (second_phase_out_end - second_phase_out_start),
            1,
        )
        return where(
            earned_income > second_phase_out_start,
            amount_after_first_phase_out
            * (1 - percentage_along_second_phase_out),
            amount_after_first_phase_out,
        )



from policyengine_us.model_api import *


class income_tax_excluding_ptc(Variable):
    value_type = float
    entity = TaxUnit
    label = "income tax (excluding PTC)"
    unit = USD
    definition_period = YEAR
    adds = ["income_tax"]
    subtracts = ["aca_ptc"]



from policyengine_us.model_api import *


class la_general_relief_housing_subsidy(Variable):
    value_type = float
    entity = SPMUnit
    label = "Los Angeles County General Relief Housing Subsidy"
    definition_period = MONTH
    # Person has to be a resident of LA County
    defined_for = "la_general_relief_housing_subsidy_eligible"
    reference = "https://dpss.lacounty.gov/en/cash/gr/housing.html"

    def formula(spm_unit, period, parameters):
        married = add(spm_unit, period, ["is_married"])
        p = parameters(period).gov.local.ca.la.general_relief.housing_subsidy
        subsidy_amount = where(married, p.amount.married, p.amount.single)
        # If filers are receiving the housing subsidy, they are obligated
        # to contribute an amount of their GR towards rent
        rent_contributions = spm_unit(
            "la_general_relief_rent_contribution", period
        )
        # The amount can not exceed rent
        rent = add(spm_unit, period, ["rent"])
        return min_(rent, subsidy_amount + rent_contributions)



from policyengine_us.model_api import *


class ma_mbta_senior_charlie_card_eligible(Variable):
    value_type = bool
    entity = Person
    label = "Eligible for the Massachusetts Bay Transportation Authority Senior Charlie Card program"
    reference = "https://www.mbta.com/fares/reduced/senior-charliecard"
    definition_period = YEAR
    defined_for = StateCode.MA

    def formula(person, period, parameters):
        age = person("age", period)
        p = parameters(period).gov.states.ma.dot.mbta.senior_charlie_card
        return age >= p.age_threshold



from policyengine_us.model_api import *


class is_ccdf_age_eligible(Variable):
    value_type = bool
    entity = Person
    definition_period = YEAR
    label = "Age eligibility for CCDF"

    def formula(person, period, parameters):
        age = person("age", period)
        age_limit = parameters(period).gov.hhs.ccdf.age_limit
        return age < age_limit



from policyengine_us.model_api import *


class va_capped_state_and_local_sales_or_income_tax(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "Capped state and local sales or income tax for Virginia itemized deductions purposes"
    unit = USD

    def formula(tax_unit, period, parameters):
        uncapped_state_and_local_tax = tax_unit(
            "state_and_local_sales_or_income_tax", period
        )
        p_salt = parameters(
            period
        ).gov.irs.deductions.itemized.salt_and_real_estate
        filing_status = tax_unit("filing_status", period)
        real_estate_tax = add(tax_unit, period, ["real_estate_taxes"])
        state_and_local_tax_cap = p_salt.cap[filing_status]
        cap_reduced_by_real_estate_tax = max_(
            state_and_local_tax_cap - real_estate_tax, 0
        )
        return min_(
            uncapped_state_and_local_tax, cap_reduced_by_real_estate_tax
        )



from policyengine_us.model_api import *


class ma_ccfa_center_based_early_education_reimbursement(Variable):
    value_type = float
    entity = Person
    label = "Massachusetts Child Care Financial Assistance (CCFA) center-based early education reimbursement amount per child"
    unit = USD
    reference = "https://www.mass.gov/doc/fiscal-year-2025-child-care-financial-assistance-daily-reimbursement-rates/download"
    definition_period = MONTH
    defined_for = StateCode.MA

    def formula(person, period, parameters):
        p = parameters(
            period
        ).gov.states.ma.eec.ccfa.reimbursement_rates.center_based
        region = person.household("ma_ccfa_region", period)
        age_category = person("ma_ccfa_child_age_category", period)
        return p.early_education[region][age_category]



from policyengine_us.model_api import *


class ctc_social_security_tax(Variable):
    value_type = float
    entity = TaxUnit
    label = "Refundable Child Tax Credit Social Security Tax"
    unit = USD
    documentation = (
        "Social Security taxes considered in the Child Tax Credit calculation"
    )
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/24#d_2"

    adds = "gov.irs.credits.ctc.refundable.social_security.add"
    subtracts = "gov.irs.credits.ctc.refundable.social_security.subtract"



from policyengine_us.model_api import *


class mo_qualified_health_insurance_premiums(Variable):
    value_type = float
    entity = Person
    label = "Missouri qualified healh insurance premiums"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://dor.mo.gov/forms/5695.pdf",  # MO Form 5695
        "https://www.irs.gov/pub/irs-pdf/f1040sa.pdf",  # US Form 1040 Sch A
        "https://www.irs.gov/pub/irs-pdf/f1040.pdf",  # US Form 1040
    )
    defined_for = StateCode.MO

    def formula(person, period, parameters):
        # Variable involves person- and tax unit-level variables.
        tax_unit = person.tax_unit

        # total_health_insurance_premiums is also a primary input to
        # the MO side of calculation, MO Form 5695, Line 8
        # Federal Schedule A, Line 4
        med_expense_deduction = add(
            tax_unit, period, ["medical_expense_deduction"]
        )

        # the ratio of federal medical expense deduction to
        # total medical expenses (out of pocket + premiums)
        # need division because med_dental_out_of_pocket is
        # in federal tax, but no MO tax
        # this ratio is then used to scale the health_insurance_premium
        # amount that can be claimed
        # IRS Schedule A Line 1
        tax_unit_health_expenses = add(
            tax_unit,
            period,
            # Out of pocket expenses include health insurance premiums
            ["medical_out_of_pocket_expenses"],
        )
        med_expense_ratio = np.zeros_like(tax_unit_health_expenses)
        mask = tax_unit_health_expenses > 0
        med_expense_ratio[mask] = (
            med_expense_deduction[mask] / tax_unit_health_expenses[mask]
        )

        person_premiums = person("health_insurance_premiums", period)
        tax_unit_premiums = tax_unit.sum(person_premiums)

        # Line 13 of MO Form 5695, represents the portion of
        # medical expenses already deducted via federal tax itemization
        deducted_portion = tax_unit_premiums * med_expense_ratio
        # subtracts the portion of premiums already deducted from federal tax
        itemized_premiums_amount = max_(
            tax_unit_premiums - deducted_portion, 0
        )
        itemizes = tax_unit("tax_unit_itemizes", period)
        # Cap at federal taxable income.
        taxable_income = tax_unit("taxable_income", period)
        # Allocate proportionally across people in the tax unit based on share
        # of premiums.
        # Use a mask rather than where to avoid a divide-by-zero warning. Default to zero.
        person_share = np.zeros_like(tax_unit_premiums)
        mask = tax_unit_premiums > 0
        person_share[mask] = person_premiums[mask] / tax_unit_premiums[mask]
        return person_share * where(
            itemizes,
            min_(itemized_premiums_amount, taxable_income),
            min_(tax_unit_premiums, taxable_income),
        )



from policyengine_us.model_api import *


class wv_social_security_benefits_subtraction(Variable):
    value_type = float
    entity = TaxUnit
    label = "West Virginia social security benefits subtraction"
    unit = USD
    definition_period = YEAR
    reference = (
        # West Virginia Personal Income Tax Forms And Instructions 2020 LINE 33
        "https://tax.wv.gov/Documents/TaxForms/2020/it140.booklet.pdf#page=24",
        # West Virginia Personal Income Tax Forms And Instructions 2020 LINE 32
        "https://tax.wv.gov/Documents/TaxForms/2021/it140.booklet.pdf#page=24",
        # West Virginia Personal Income Tax Forms And Instructions 2022 LINE 32
        "https://tax.wv.gov/Documents/PIT/2022/PersonalIncomeTaxFormsAndInstructions.2022.pdf#page=25",
        # Code of West Virginia §11-21-12 (c)(8)(A) - (c)(8)(F)
        "https://code.wvlegislature.gov/11-21-12/",
    )
    defined_for = StateCode.WV

    adds = ["wv_social_security_benefits_subtraction_person"]



from policyengine_us.model_api import *


class co_tabor_cash_back(Variable):
    value_type = float
    entity = TaxUnit
    label = "Colorado TABOR cash back"
    defined_for = StateCode.CO
    unit = USD
    definition_period = YEAR
    reference = "https://leg.colorado.gov/sites/default/files/documents/2022A/bills/2022a_233_01.pdf"

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.co.tax.income.credits.tabor

        joint = tax_unit("tax_unit_is_joint", period)
        return where(joint, p.joint, p.single)



from policyengine_us.model_api import *


class mo_tanf_income_limit(Variable):
    value_type = float
    entity = SPMUnit
    label = "Missouri TANF income limit / maximum benefit"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MO

    def formula(spm_unit, period, parameters):
        people = spm_unit("spm_unit_size", period)
        p = parameters(period).gov.states.mo.dss.tanf
        # TODO: Find the amount for households with more than 8 people.
        capped_people = min_(people, 8).astype(int)
        # Note: Missouri defines income limits annually.
        return p.income_limit[capped_people]



from policyengine_us.model_api import *


class ca_calworks_child_care_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Eligible for the California CalWORKs Child Care"
    definition_period = YEAR
    defined_for = StateCode.CA
    reference = "http://epolicy.dpss.lacounty.gov/epolicy/epolicy/server/general/projects_responsive/ePolicyMaster/index.htm?&area=general&type=responsivehelp&ctxid=&project=ePolicyMaster#t=mergedProjects%2FChild%20Care%2FChild_Care%2F1210_Overview%2F1210_Overview.htm%23Backgroundbc-3&rhtocid=_3_3_0_2"

    def formula(spm_unit, period, parameters):
        receives_tanf = spm_unit("ca_tanf", period) > 0
        age_eligible = spm_unit("ca_calworks_child_care_age_eligible", period)
        work_requirement = spm_unit(
            "ca_calworks_child_care_meets_work_requirement", period
        )
        person = spm_unit.members
        immigration_eligible = spm_unit.all(
            person(
                "ca_calworks_child_care_immigration_status_eligible_person",
                period,
            )
        )
        return (
            receives_tanf
            & age_eligible
            & work_requirement
            & immigration_eligible
        )



from policyengine_us.model_api import *


class il_tanf_immigration_status_eligible_person(Variable):
    value_type = bool
    entity = Person
    label = "Eligible person for the Illinois TANF based on immigration status"
    definition_period = MONTH
    reference = "https://www.law.cornell.edu/regulations/illinois/Ill-Admin-Code-tit-89-SS-112.10"
    defined_for = StateCode.IL

    def formula(person, period, parameters):
        p = parameters(period).gov.states.il.dhs.tanf
        immigration_status = person("immigration_status", period)
        immigration_status_str = immigration_status.decode_to_str()
        has_qualifying_status = np.isin(
            immigration_status_str,
            p.qualified_noncitizen_statuses,
        )
        is_citizen = (
            immigration_status == immigration_status.possible_values.CITIZEN
        )

        return has_qualifying_status | is_citizen



from policyengine_us.model_api import *


class ctc_arpa_phase_out_cap(Variable):
    value_type = float
    entity = TaxUnit
    label = "Cap on phase-out of ARPA CTC expansion"
    unit = USD
    documentation = (
        "ARPA capped the additional amount based on the phase-out thresholds."
    )
    definition_period = YEAR
    # Defined on Line 5 worksheet of 2021 Instructions for Schedule 8812.
    reference = "https://www.irs.gov/pub/irs-pdf/i1040s8.pdf#page=4"

    def formula(tax_unit, period, parameters):
        # Logic sequence follows the form, which is clearer than the IRC.
        p = parameters(period).gov.irs.credits.ctc
        # defined_for didn't work.
        if not p.phase_out.arpa.in_effect:
            return 0
        arpa_threshold = tax_unit("ctc_arpa_phase_out_threshold", period)
        original_threshold = tax_unit("ctc_phase_out_threshold", period)
        threshold_diff = original_threshold - arpa_threshold
        cap = (
            threshold_diff
            * p.amount.arpa_expansion_cap_percent_of_threshold_diff
        )
        # Form limits the cap to the amount of the increase in the maximum CTC.
        return min_(cap, tax_unit("ctc_arpa_max_addition", period))



from policyengine_us.model_api import *


class va_income_tax_if_claiming_non_refundable_eitc(Variable):
    value_type = float
    entity = TaxUnit
    label = "Virginia tax liability if claiming non-refundable Virginia EITC"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.VA

    def formula(tax_unit, period, parameters):
        simulation = tax_unit.simulation
        non_refundable_branch = simulation.get_branch("va_non_refundable_eitc")
        non_refundable_branch.set_input(
            "va_claims_refundable_eitc",
            period,
            np.zeros((tax_unit.count,), dtype=bool),
        )
        return non_refundable_branch.calculate("va_income_tax", period)



from policyengine_us.model_api import *


class able_contributions_person(Variable):
    value_type = float
    entity = Person
    label = "Person-level ABLE contributions"
    documentation = "Contributions made to an ABLE account by each individual."
    unit = USD
    definition_period = YEAR



from policyengine_us.model_api import *
from policyengine_core.periods import period as period_
from policyengine_core.periods import instant


def create_additional_tax_bracket() -> Reform:
    class regular_tax_before_credits(Variable):
        value_type = float
        entity = TaxUnit
        definition_period = YEAR
        label = "Regular tax before credits"
        documentation = "Regular tax on regular taxable income before credits"
        unit = USD

        def formula(tax_unit, period, parameters):
            p = parameters(period).gov.irs
            filing_status = tax_unit("filing_status", period)
            dwks1 = tax_unit("taxable_income", period)

            dwks16 = min_(
                p.capital_gains.thresholds["1"][filing_status], dwks1
            )
            dwks17 = min_(tax_unit("dwks14", period), dwks16)
            dwks20 = dwks16 - dwks17
            lowest_rate_tax = p.capital_gains.rates["1"] * dwks20
            # Break in worksheet lines
            dwks13 = tax_unit("dwks13", period)
            dwks21 = min_(dwks1, dwks13)
            dwks22 = dwks20
            dwks23 = max_(0, dwks21 - dwks22)
            dwks25 = min_(
                p.capital_gains.thresholds["2"][filing_status], dwks1
            )
            dwks19 = tax_unit("dwks19", period)
            dwks26 = min_(dwks19, dwks20)
            dwks27 = max_(0, dwks25 - dwks26)
            dwks28 = min_(dwks23, dwks27)
            dwks29 = p.capital_gains.rates["2"] * dwks28
            dwks30 = dwks22 + dwks28
            dwks31 = dwks21 - dwks30
            dwks32 = p.capital_gains.rates["3"] * dwks31
            # Break in worksheet lines
            dwks33 = min_(
                tax_unit("dwks09", period),
                add(tax_unit, period, ["unrecaptured_section_1250_gain"]),
            )
            dwks10 = tax_unit("dwks10", period)
            dwks34 = dwks10 + dwks19
            dwks36 = max_(0, dwks34 - dwks1)
            dwks37 = max_(0, dwks33 - dwks36)

            p = parameters(period).gov.irs.income

            dwks38 = p.amt.capital_gains.capital_gain_excess_tax_rate * dwks37
            # Break in worksheet lines
            dwks39 = dwks19 + dwks20 + dwks28 + dwks31 + dwks37
            dwks40 = dwks1 - dwks39
            dwks41 = p.amt.brackets.rates[-1] * dwks40

            # Compute regular tax using bracket rates and thresholds
            reg_taxinc = max_(0, dwks19)
            p_reform = parameters(period).gov.contrib.additional_tax_bracket
            bracket_tops = p_reform.bracket.thresholds
            bracket_rates = p_reform.bracket.rates
            reg_tax = 0
            bracket_bottom = 0
            for i in range(1, len(list(bracket_rates.__iter__())) + 1):
                b = str(i)
                bracket_top = bracket_tops[b][filing_status]
                reg_tax += bracket_rates[b] * amount_between(
                    reg_taxinc, bracket_bottom, bracket_top
                )
                bracket_bottom = bracket_top

            # Return to worksheet lines
            dwks42 = reg_tax
            dwks43 = (
                dwks29 + dwks32 + dwks38 + dwks41 + dwks42 + lowest_rate_tax
            )
            dwks44 = tax_unit("income_tax_main_rates", period)
            dwks45 = min_(dwks43, dwks44)
            return where(tax_unit("has_qdiv_or_ltcg", period), dwks45, dwks44)

    class income_tax_main_rates(Variable):
        value_type = float
        entity = TaxUnit
        definition_period = YEAR
        label = "Income tax main rates"
        reference = "https://www.law.cornell.edu/uscode/text/26/1"
        unit = USD

        def formula(tax_unit, period, parameters):
            # compute taxable income that is taxed at the main rates
            full_taxable_income = tax_unit("taxable_income", period)
            cg_exclusion = tax_unit(
                "capital_gains_excluded_from_taxable_income", period
            )
            taxinc = max_(0, full_taxable_income - cg_exclusion)
            # compute tax using bracket rates and thresholds
            p = parameters(period).gov.contrib.additional_tax_bracket
            bracket_tops = p.bracket.thresholds
            bracket_rates = p.bracket.rates
            filing_status = tax_unit("filing_status", period)
            tax = 0
            bracket_bottom = 0
            for i in range(1, len(list(bracket_rates.__iter__())) + 1):
                b = str(i)
                bracket_top = bracket_tops[b][filing_status]
                tax += bracket_rates[b] * amount_between(
                    taxinc, bracket_bottom, bracket_top
                )
                bracket_bottom = bracket_top
            return tax

    class reform(Reform):
        def apply(self):
            self.update_variable(income_tax_main_rates)
            self.update_variable(regular_tax_before_credits)

    return reform


def create_additional_tax_bracket_reform(
    parameters, period, bypass: bool = False
):
    if bypass:
        return create_additional_tax_bracket()

    p = parameters.gov.contrib.additional_tax_bracket

    reform_active = False
    current_period = period_(period)

    for i in range(5):
        if p(current_period).in_effect:
            reform_active = True
            break
        current_period = current_period.offset(1, "year")

    if reform_active:
        return create_additional_tax_bracket()
    else:
        return None


additional_tax_bracket = create_additional_tax_bracket_reform(
    None, None, bypass=True
)



from policyengine_us.model_api import *


class ca_ala_general_assistance(Variable):
    value_type = float
    entity = SPMUnit
    unit = USD
    label = "Alameda County General Assistance base amount"
    definition_period = MONTH
    defined_for = "ca_ala_general_assistance_income_eligible"
    reference = "https://www.alamedacountysocialservices.org/acssa-assets/PDF/GA-Policies/GA-Regulations.pdf#page=28"

    # Defined for income eligible family, which make sure that countable income is less than based amount
    adds = ["ca_ala_general_assistance_base_amount"]
    subtracts = ["ca_ala_general_assistance_countable_income"]



from policyengine_us.model_api import *


class mt_elderly_homeowner_or_renter_credit(Variable):
    value_type = float
    entity = Person
    label = "Montana Elderly Homeowner/Renter Credit"
    unit = USD
    definition_period = YEAR
    defined_for = "mt_elderly_homeowner_or_renter_credit_eligible"

    def formula(person, period, parameters):
        p = parameters(
            period
        ).gov.states.mt.tax.income.credits.elderly_homeowner_or_renter
        gross_household_income = person(
            "mt_elderly_homeowner_or_renter_credit_gross_household_income",
            period,
        )
        # Get net_household_income and allocate it to the head
        head = person("is_tax_unit_head", period)
        net_household_income = add(
            person.tax_unit,
            period,
            ["mt_elderly_homeowner_or_renter_credit_net_household_income"],
        )
        # Credit Computation
        property_tax = add(person.tax_unit, period, ["real_estate_taxes"])
        rent = add(person.tax_unit, period, ["rent"])
        countable_rent = rent * p.rent_equivalent_tax_rate
        countable_rent_and_property_tax = property_tax + countable_rent
        uncapped_credit_unit = max_(
            countable_rent_and_property_tax - net_household_income, 0
        )
        uncapped_credit = uncapped_credit_unit * head
        capped_credit = min_(uncapped_credit, p.cap)
        multiplier = p.multiplier.calc(gross_household_income)
        return capped_credit * multiplier



from policyengine_us.model_api import *


class me_property_tax_fairness_credit_countable_rent(Variable):
    value_type = float
    entity = TaxUnit
    unit = USD
    label = "Countable rent for Maine property tax fairness credit"
    definition_period = YEAR
    defined_for = StateCode.ME
    reference = "https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_sched_pstfc_ff.pdf#page=2"

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.me.tax.income.credits.fairness.property_tax
        rent = add(tax_unit, period, ["rent"])
        utilities_included_in_rent = tax_unit(
            "utilities_included_in_rent", period
        )
        utility_expenses = add(tax_unit, period, ["utility_expense"])
        # A separate calculation exists for the case where utilities are included in rent
        # if the filer does not know the portion of rent that is attributable to utilities
        # This is not implemented
        deductible_utility_expenses = (
            utilities_included_in_rent * utility_expenses
        )
        net_rent = rent - deductible_utility_expenses
        return net_rent * p.rate.rent



from policyengine_us.model_api import *


class ga_deductions(Variable):
    value_type = float
    entity = TaxUnit
    label = "Georgia deductions"
    unit = USD
    reference = (
        # (a)(1) ties itemization to the federal choice.
        "https://law.justia.com/codes/georgia/2022/title-48/chapter-7/article-2/section-48-7-27/",
        # Page 12 states:
        # "If you use the standard deduction on your Federal
        #  return, you must use the Georgia standard deduction on
        #  your Georgia return."
        # Page 13 states:
        # "If you itemize deductions on your Federal return, or if you
        #  are married filing separate and your spouse itemizes deductions,
        #  you must itemize deductions on your Georgia return."
        "https://dor.georgia.gov/document/document/2022-it-511-individual-income-tax-booklet/download",
    )
    definition_period = YEAR
    defined_for = StateCode.GA

    def formula(tax_unit, period, parameters):
        itemizes = tax_unit("tax_unit_itemizes", period)
        sd = tax_unit("ga_standard_deduction", period)
        # 48-7-27(a)(1) states:
        # "Either the sum of all itemized nonbusiness deductions used in computing
        #  such taxpayer’s federal taxable income or..."
        p = parameters(period).gov.irs.deductions
        itemized = add(tax_unit, period, p.itemized_deductions)
        return where(itemizes, itemized, sd)



from policyengine_us.model_api import *


class or_severely_disabled_exemptions(Variable):
    value_type = int
    entity = TaxUnit
    label = "OR severely disabled exemptions for tax head or spouse"
    definition_period = YEAR
    reference = (
        "https://www.oregon.gov/dor/forms/FormsPubs/form-or-40-inst_101-040-1_2021.pdf#page=17"
        "https://www.oregonlegislature.gov/bills_laws/ors/ors316.html"  # Subsection 316.758(1)(b)
    )
    defined_for = StateCode.OR

    def formula(tax_unit, period, parameters):
        # Identify qualifying tax unit members.
        person = tax_unit.members
        severely_disabled = person("is_severely_disabled", period)
        head_or_spouse = ~person("is_tax_unit_dependent", period)
        # Determine eligibility from AGI.
        federal_agi = tax_unit("adjusted_gross_income", period)
        p = parameters(period).gov.states["or"].tax.income.credits.exemption
        qualifies = federal_agi <= p.income_limit.severely_disabled
        return qualifies * tax_unit.sum(severely_disabled & head_or_spouse)



from policyengine_us.model_api import *


class az_parents_grandparents_exemption(Variable):
    value_type = float
    entity = Person
    label = "Arizona parents and grandparents exemption"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.AZ

    def formula(person, period, parameters):
        p = parameters(
            period
        ).gov.states.az.tax.income.exemptions.parent_grandparent

        parent = person("is_parent_of_filer_or_spouse", period)
        grandparent = person("is_grandparent_of_filer_or_spouse", period)

        age = person("age", period)
        age_eligible = age >= p.min_age

        ratio = person(
            "share_of_care_and_support_costs_paid_by_tax_filer", period
        )
        ratio_eligible = ratio >= p.cost_rate

        eligible_parent_or_grandparent = (
            (parent | grandparent) & age_eligible & ratio_eligible
        )

        return p.amount * eligible_parent_or_grandparent



from policyengine_us.model_api import *


class school_meal_countable_income(Variable):
    value_type = float
    entity = SPMUnit
    definition_period = YEAR
    unit = USD
    label = "Countable income for school meals"
    documentation = "SPM unit's countable income for school meal program"

    adds = "gov.usda.school_meals.income.sources"



from policyengine_us.model_api import *


class ri_child_tax_rebate_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Rhode Island Child Tax Rebate"
    definition_period = YEAR
    reference = "https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-08/H7123Aaa_CTR_0.pdf"
    defined_for = StateCode.RI

    def formula(tax_unit, period, parameters):
        income = tax_unit("adjusted_gross_income", period)
        filing_status = tax_unit("filing_status", period)
        p = parameters(
            period
        ).gov.states.ri.tax.income.credits.child_tax_rebate.limit
        return income <= p.income[filing_status]



from policyengine_us.model_api import *


class il_aged_blind_exemption(Variable):
    value_type = float
    entity = TaxUnit
    label = "IL aged and blind exemption"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.IL

    def formula(tax_unit, period, parameters):
        aged_blind_count = tax_unit("aged_blind_count", period)
        p = parameters(period).gov.states.il.tax.income.exemption
        return aged_blind_count * p.aged_and_blind



from policyengine_us.model_api import *


class il_hbwd_person(Variable):
    value_type = float
    entity = Person
    label = "Illinois Health Benefits for Workers with Disabilities per person"
    unit = USD
    definition_period = MONTH
    reference = (
        "https://hfs.illinois.gov/medicalprograms/hbwd.html",
        "https://www.law.cornell.edu/regulations/illinois/Ill-Admin-Code-tit-89-SS-120.510",
    )
    defined_for = "il_hbwd_eligible"

    def formula(person, period, parameters):
        # HBWD provides comprehensive healthcare coverage
        # Could model as:
        # 1. Negative premium (cost to individual)
        # 2. Imputed healthcare value minus premium
        # 3. Healthcare coverage indicator
        # For now, return negative premium as cost
        premium = person("il_hbwd_premium", period)
        return -premium



from policyengine_us.model_api import *


class pell_grant_dependent_other_allowances(Variable):
    value_type = float
    entity = Person
    unit = USD
    label = "Pell Grant dependent other allowances"
    definition_period = YEAR



from policyengine_us.model_api import *


class snap_emergency_allotment(Variable):
    value_type = float
    entity = SPMUnit
    definition_period = MONTH
    documentation = "SNAP emergency allotment"
    label = "SNAP emergency allotment"
    reference = "https://www.law.cornell.edu/uscode/text/7/2017#a"
    unit = USD

    def formula(spm_unit, period, parameters):
        # Save time by only recalculating the EA parameters for each month,
        # rather than the whole parameter tree for each month.
        p = parameters.gov.usda.snap.emergency_allotment(period)
        if not p.allowed:
            return 0
        # Federal SNAP rules are defined in U.S.C Title 7, Chapter 51, which also
        # defines state powers to modify the rules.
        eligible = spm_unit("is_snap_eligible", period)
        max_allotment = spm_unit("snap_max_allotment", period)
        normal_allotment = spm_unit("snap_normal_allotment", period)
        # Calculate emergency allotment, which provides all eligible households the maximum.
        state = spm_unit.household("state_code", period)
        ea_in_effect = p.in_effect[state]
        ea_minimum = p.minimum
        ea_amount_if_in_effect = max_(
            ea_minimum, max_allotment - normal_allotment
        )
        return eligible * ea_in_effect * ea_amount_if_in_effect



from policyengine_us.model_api import *


class mt_taxable_income_indiv(Variable):
    value_type = float
    entity = Person
    label = "Montana taxable income when married couples are filing separately"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MT
    reference = (
        "https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/05/Montana-Idividiual-Income-Tax-Return-Form-2-2022v6.2.pdf#page=1",
        "https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2022/12/Form-2-2022-Instructions.pdf#page=16",
    )

    def formula(person, period, parameters):
        pre_dependent_exemption_income = person(
            "mt_pre_dependent_exemption_taxable_income_indiv", period
        )
        dependent_exemptions = add(
            person.tax_unit, period, ["mt_dependent_exemptions_person"]
        )
        # The dependent exemptions are allocated optimally between the head and spouse
        head = person("is_tax_unit_head", period)
        spouse = person("is_tax_unit_spouse", period)
        head_income = person.tax_unit.sum(
            pre_dependent_exemption_income * head
        )
        spouse_income = person.tax_unit.sum(
            pre_dependent_exemption_income * spouse
        )
        # Calculate the difference between the head and spouse income
        income_difference = np.abs(head_income - spouse_income)
        # Cap the exemption amount initially at the difference between the head and spouse income
        exemptions_capped_at_income_difference = min_(
            income_difference, dependent_exemptions
        )
        head_income_exceeds_spouse_income = head_income > spouse_income
        # The capped exemptions are allocated to the spouse with the larger income
        spouse_with_larger_income = where(
            head_income_exceeds_spouse_income, head, spouse
        )
        exemptions_capped_at_income_difference_allocated_to_spouse_with_larger_income = (
            exemptions_capped_at_income_difference * spouse_with_larger_income
        )
        # The remaining exemption amount is halved and allocated between the head and spouse
        remaining_exemptions = (
            dependent_exemptions - exemptions_capped_at_income_difference
        )
        halved_remaining_exemptions = remaining_exemptions / 2
        return max_(
            pre_dependent_exemption_income
            - exemptions_capped_at_income_difference_allocated_to_spouse_with_larger_income
            - halved_remaining_exemptions,
            0,
        )



from policyengine_us.model_api import *


class taxsim_pui(Variable):
    value_type = float
    entity = TaxUnit
    label = "Unemployment compensation (primary taxpayer)"
    unit = USD
    documentation = "Unemployment Compensation received - primary taxpayer."
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        person = tax_unit.members
        is_primary = person("is_tax_unit_head", period)
        unemployment_compensation = person("unemployment_compensation", period)
        return tax_unit.sum(unemployment_compensation * is_primary)



from policyengine_us.model_api import *


class is_tanf_non_cash_hheod(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Elderly or disabled for TANF non-cash benefit"
    documentation = "Whether the household is considered elderly or disabled for TANF non-cash benefit for SNAP BBCE"
    definition_period = MONTH

    def formula(spm_unit, period, parameters):
        state = spm_unit.household("state_code_str", period)
        bbce = parameters(period).gov.hhs.tanf.non_cash
        requires_all = bbce.requires_all_for_hheod[state]
        return where(
            requires_all,
            spm_unit("has_all_usda_elderly_disabled", period),
            spm_unit("has_usda_elderly_disabled", period),
        )



from policyengine_us.model_api import *


class ri_exemptions(Variable):
    value_type = float
    entity = TaxUnit
    label = "Rhode Island exemptions"
    unit = USD
    definition_period = YEAR
    reference = "https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022%20Tax%20Rate%20and%20Worksheets.pdf"
    defined_for = StateCode.RI

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.ri.tax.income.exemption

        exemptions_count = tax_unit("exemptions_count", period)

        exemption_amount = exemptions_count * p.amount

        # Modified Federal AGI
        mod_agi = tax_unit("ri_agi", period)

        excess_agi = max_(0, mod_agi - p.reduction.start)

        increments = np.ceil(excess_agi / p.reduction.increment)

        reduction_rate = min_(p.reduction.rate * increments, 1)

        return exemption_amount * (1 - reduction_rate)



from policyengine_us.model_api import *


class ma_eaedc_assets_limit_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Eligible based on the asset limit for the Massachusetts EAEDC"
    definition_period = MONTH
    defined_for = StateCode.MA
    reference = (
        "https://www.law.cornell.edu/regulations/massachusetts/106-CMR-704-110"
    )

    def formula(spm_unit, period, parameters):
        p = parameters(period).gov.states.ma.dta.tcap.eaedc.assets
        countable_assets = spm_unit("ma_eaedc_countable_assets", period)
        living_arrangement = spm_unit("ma_eaedc_living_arrangement", period)

        living_arrangement_E = (
            living_arrangement == living_arrangement.possible_values.E
        )
        assets_below_limit = countable_assets <= p.limit
        # Only living arrangement E has an asset limit

        return (
            living_arrangement_E & assets_below_limit
        ) | ~living_arrangement_E



from policyengine_us.model_api import *


class nm_deduction_for_certain_dependents_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Eligibility for New Mexico deduction for certain dependents"
    definition_period = YEAR
    reference = "https://nmonesource.com/nmos/nmsa/en/item/4340/index.do#!fragment/zoupio-_Toc140503892/BQCwhgziBcwMYgK4DsDWszIQewE4BUBTADwBdoAvbRABwEtsBaAfX2zgEYAWABgFYeAZgAcATgBMASgA0ybKUIQAiokK4AntADkW6REJhcCFWs069BoyADKeUgCFNAJQCiAGRcA1AIIA5AMIu0qRgAEbQpOySkkA"
    defined_for = StateCode.NM

    def formula(tax_unit, period, parameters):
        dependent_elsewhere = tax_unit("head_is_dependent_elsewhere", period)
        # deduction does not apply if an exemption under IRS 151 is claimed;
        # IRC 151 refers to the federal personal exemption
        federal_exemption_amount = tax_unit("exemptions", period)
        return ~dependent_elsewhere & (federal_exemption_amount == 0)



from policyengine_us.model_api import *


class rent_is_shared_with_another_tax_unit(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Whether the household shares rent with others"
    definition_period = YEAR



from policyengine_us.model_api import *


class filer_meets_ctc_identification_requirements(Variable):
    value_type = bool
    entity = TaxUnit
    definition_period = YEAR
    label = "Filer meets CTC identification requirements"

    def formula(tax_unit, period, parameters):
        # Both head and spouse in the tax unit must have valid SSN card type to be eligible for the CTC
        p = parameters(period).gov.irs.credits.ctc
        if p.adult_ssn_requirement_applies:
            person = tax_unit.members
            head_or_spouse = person("is_tax_unit_head_or_spouse", period)
            eligible_ssn_card_type = person(
                "meets_ctc_identification_requirements", period
            )
            return tax_unit.any(head_or_spouse & eligible_ssn_card_type)
        return True



from policyengine_us.model_api import *


class mi_retirement_benefits_deduction_tier_three_ss_exempt_not_retired(
    Variable
):
    value_type = float
    entity = TaxUnit
    label = "Michigan non-retired tier three retirement benefits deduction"
    unit = USD
    definition_period = YEAR
    reference = (
        "http://legislature.mi.gov/doc.aspx?mcl-206-30",  # (9)(d)
        "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf#page=17",
        "https://www.michigan.gov/taxes/iit/retirement-and-pension-benefits",
        "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/4884.pdf",
        "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/Form-4884-Section-C-worksheet.pdf",
    )
    defined_for = "mi_retirement_benefits_deduction_tier_three_eligible"

    def formula(tax_unit, period, parameters):
        # Modeled after 2022 Michigan Pension Schedule (Form 4884) Section C
        p = parameters(
            period
        ).gov.states.mi.tax.income.deductions.retirement_benefits
        #  Recipients should receive retirement benefits from employment exempt from Social Security
        eligible_people = tax_unit(
            "mi_retirement_benefits_deduction_tier_three_ss_exempt_not_retired_eligible_people",
            period,
        )

        person = tax_unit.members
        uncapped_pension_income = person("taxable_pension_income", period)
        is_head_or_spouse = person("is_tax_unit_head_or_spouse", period)

        # Head and spouse both are eligible to receive an equal deduction amount
        cap = (
            p.tier_three.ss_exempt.not_retired.amount * eligible_people
        )  # Line 9
        uncapped_head_or_spouse_pension = tax_unit.sum(
            uncapped_pension_income * is_head_or_spouse
        )

        # If a filer is recieving military retirement pay, the calculation includes the smaller of
        # the tier one or tier three deduction amount
        military_retirement_pay_received = (
            tax_unit.sum(person("military_retirement_pay", period)) > 0
        )
        # Line 8
        tier_one_amount = tax_unit(
            "mi_retirement_benefits_deduction_tier_one_amount",
            period,
        )
        # Line 10
        smaller_of_cap_or_tier_one_amount = min_(cap, tier_one_amount)

        eligible_deduction = where(
            military_retirement_pay_received,
            smaller_of_cap_or_tier_one_amount,
            cap,
        )
        # Line 18
        tier_three_ss_exempt_not_retired = min_(
            uncapped_head_or_spouse_pension, eligible_deduction
        )
        # Worksheet 3.3: Retirement and Pension Benefits Subtraction for Section D of Form 4884
        if p.expanded.availability:
            expanded_retirement_benefits_deduction = tax_unit(
                "mi_expanded_retirement_benefits_deduction",
                period,
            )
            return max_(
                tier_three_ss_exempt_not_retired,
                expanded_retirement_benefits_deduction,
            )
        else:
            return tier_three_ss_exempt_not_retired



from policyengine_us.model_api import *


class ms_cdcc(Variable):
    value_type = float
    entity = TaxUnit
    label = "Mississippi child and dependent care credit"
    unit = USD
    definition_period = YEAR
    defined_for = "ms_cdcc_eligible"
    reference = "https://legiscan.com/MS/text/HB1671/id/2767768"

    def formula(tax_unit, period, parameters):
        # Mississippi matches the federal credit taken
        cdcc = tax_unit("cdcc", period)
        p = parameters(period).gov.states.ms.tax.income.credits.cdcc
        return p.match * cdcc



from policyengine_us.model_api import *


class min_head_spouse_earned(Variable):
    value_type = float
    entity = TaxUnit
    label = "Less of head and spouse's earnings"
    unit = USD
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        is_joint = tax_unit("tax_unit_is_joint", period)
        head_earnings = tax_unit("head_earned", period)
        spouse_earnings = tax_unit("spouse_earned", period)
        return where(
            is_joint, min_(head_earnings, spouse_earnings), head_earnings
        )



from policyengine_us.model_api import *


class tax_liability_if_itemizing(Variable):
    value_type = float
    entity = TaxUnit
    label = "Tax liability if itemizing"
    unit = USD
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        simulation = tax_unit.simulation
        itemized_branch = simulation.get_branch("itemizing")
        itemized_branch.set_input(
            "tax_unit_itemizes", period, np.ones((tax_unit.count,), dtype=bool)
        )
        return itemized_branch.calculate("income_tax", period)



from policyengine_us.model_api import *


class ks_income_tax_before_refundable_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "Kansas income tax before refundable credits"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.ksrevenue.gov/pdf/ip21.pdf"
        "https://www.ksrevenue.gov/pdf/ip22.pdf"
    )
    defined_for = StateCode.KS

    def formula(tax_unit, period, parameters):
        itax_before_credits = tax_unit("ks_income_tax_before_credits", period)
        nonrefundable_credits = tax_unit("ks_non_refundable_credits", period)
        return max_(0, itax_before_credits - nonrefundable_credits)



from policyengine_us.model_api import *


class co_tanf_countable_earned_income_grant_standard(Variable):
    value_type = float
    entity = SPMUnit
    label = "Colorado TANF total countable earned income for grant standard"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.CO

    def formula(spm_unit, period, parameters):
        gross_earnings = spm_unit(
            "co_tanf_countable_gross_earned_income", period
        )
        p = parameters(period).gov.states.co.cdhs.tanf.income.earned_exclusion
        # Grant standard applies a percent exclusion.
        return gross_earnings * (1 - p.percent)



from policyengine_us.model_api import *


class pr_casualty_loss_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Puerto Rico casualty loss deduction"
    unit = USD
    definition_period = YEAR
    reference = "https://law.justia.com/codes/puerto-rico/title-thirteen/subtitle-17/part-ii/chapter-1005/subchapter-c/30135/"  # (10)
    defined_for = StateCode.PR

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.territories.pr.tax.income.taxable_income.deductions.casualty_loss
        loss = add(tax_unit, period, ["casualty_loss"])
        filing_status = tax_unit("filing_status", period)

        return (
            where(
                filing_status == filing_status.possible_values.SEPARATE,
                p.separate_percentage,
                1,
            )
            * loss
        )



from policyengine_us.model_api import *


class ga_retirement_exclusion_person(Variable):
    value_type = float
    entity = Person
    label = "Georgia retirement exclusion for each person"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://dor.georgia.gov/document/booklet/2021-it-511-individual-income-tax-booklet/download",  # Page 15
        "https://dor.georgia.gov/document/document/2022-it-511-individual-income-tax-booklet/download",  # Page 15
        "https://advance.lexis.com/documentpage/?pdmfid=1000516&crid=fb5db531-a80f-4790-bddb-eefc8327ef60&config=00JAA1MDBlYzczZi1lYjFlLTQxMTgtYWE3OS02YTgyOGM2NWJlMDYKAFBvZENhdGFsb2feed0oM9qoQOMCSJFX5qkd&pddocfullpath=%2Fshared%2Fdocument%2Fstatutes-legislation%2Furn%3AcontentItem%3A65D2-CDH3-CGX8-044N-00008-00&pdcontentcomponentid=234186&pdteaserkey=sr1&pditab=allpods&ecomp=8s65kkk&earg=sr1&prid=66f02b0a-c5ae-4162-9535-127751546807",
    )
    defined_for = "ga_retirement_exclusion_eligible_person"

    def formula(person, period, parameters):
        p = parameters(
            period
        ).gov.states.ga.tax.income.agi.exclusions.retirement
        older_age_threshold_cap_eligible = (
            person("age", period) >= p.age_threshold.older
        )
        cap = where(
            older_age_threshold_cap_eligible, p.cap.older, p.cap.younger
        )
        retirement_income = person(
            "ga_retirement_income_exclusion_retirement_income", period
        )
        return min_(retirement_income, cap)



from policyengine_us.model_api import *


class ct_pension_annuity_subtraction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Connecticut pension and annuity subtraction"
    unit = USD
    reference = (
        "https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-701",  # Sec. 12-701 (20)(B)(xxi)&(xxii)
        "https://portal.ct.gov/-/media/drs/forms/2024/income/2024-ct-1040-instructions_1224.pdf#page=28",
    )
    definition_period = YEAR
    defined_for = StateCode.CT

    def formula(tax_unit, period, parameters):
        person = tax_unit.members
        filing_status = tax_unit("filing_status", period)
        status = filing_status.possible_values
        head_or_spouse = person("is_tax_unit_head_or_spouse", period)
        agi = tax_unit("adjusted_gross_income", period)

        p = parameters(
            period
        ).gov.states.ct.tax.income.subtractions.pensions_or_annuity

        # Get the rate based on AGI and filing status
        is_joint = (filing_status == status.JOINT) | (
            filing_status == status.SURVIVING_SPOUSE
        )

        rate = where(is_joint, p.joint.calc(agi), p.non_joint.calc(agi))

        # Apply the rate to eligible pension income
        pension_income = person("taxable_pension_income", period)
        eligible_pension = pension_income * head_or_spouse
        total_pension = tax_unit.sum(eligible_pension)

        return total_pension * rate



from policyengine_us.model_api import *


class ma_ccfa_activity_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Eligible for Massachusetts Child Care Financial Assistance (CCFA) due to activity"
    definition_period = MONTH
    defined_for = StateCode.MA
    reference = "https://regulations.justia.com/states/massachusetts/606-cmr/title-606-cmr-10-00/section-10-04/"

    def formula(spm_unit, period, parameters):
        p = parameters(period).gov.states.ma.eec.ccfa.activity_requirements
        person = spm_unit.members
        is_head_or_spouse = person("is_tax_unit_head_or_spouse", period)

        # Individual activity requirements
        hours_worked = person(
            "weekly_hours_worked_before_lsr", period.this_year
        )
        meets_work_requirement = hours_worked >= p.weekly_hours

        age = person("monthly_age", period)
        work_exempt_age = age >= p.work_exempt_age

        is_student = person("is_full_time_student", period)

        individually_eligible = (
            meets_work_requirement | work_exempt_age | is_student
        )

        # Family-level exemptions
        is_pregnant = person("is_pregnant", period)
        is_disabled = person("is_disabled", period)
        parent_exempt = is_head_or_spouse & (is_pregnant | is_disabled)
        family_exempt = spm_unit.any(parent_exempt)

        is_homeless = spm_unit.household("is_homeless", period)

        # All parents must meet requirements unless family is exempt
        all_parents_eligible = (
            spm_unit.sum(is_head_or_spouse & ~individually_eligible) == 0
        )

        return all_parents_eligible | family_exempt | is_homeless



from policyengine_us.model_api import *


class nc_tanf(Variable):
    value_type = float
    entity = SPMUnit
    label = "North Carolina TANF"
    unit = USD
    definition_period = YEAR
    defined_for = "nc_tanf_eligible"

    def formula(spm_unit, period, parameters):
        payment_percentage = parameters(
            period
        ).gov.states.nc.ncdhhs.tanf.need_standard.payment_percentage
        reduced_need_standard = spm_unit(
            "nc_tanf_reduced_need_standard", period
        )

        return reduced_need_standard * payment_percentage



from policyengine_us.model_api import *


class id_2022_rebate_floor(Variable):
    value_type = float
    entity = Person
    label = "Idaho 2022 rebate floor"
    definition_period = YEAR
    defined_for = StateCode.ID

    adds = ["gov.states.id.tax.income.credits.2022_rebate.floor"]



from policyengine_us.model_api import *


class cliff_evaluated(Variable):
    value_type = bool
    entity = Person
    label = "cliff evaluated"
    documentation = "Whether this person's cliff has been simulated. If not, the cliff gap is assumed to be zero."
    definition_period = YEAR

    def formula(person, period, parameters):
        adult_index_values = person("adult_index", period)
        mtr_adult_count = parameters(
            period
        ).simulation.marginal_tax_rate_adults
        is_adult = person("is_adult", period)
        return is_adult & (adult_index_values <= mtr_adult_count)


class cliff_gap(Variable):
    value_type = float
    entity = Person
    label = "cliff gap"
    unit = USD
    documentation = "Amount of income lost if this person's employment income increased by delta amount."
    definition_period = YEAR

    def formula(person, period, parameters):
        delta = parameters(period).simulation.marginal_tax_rate_delta
        mtr = person("marginal_tax_rate", period)
        return max_(0, (mtr - 1) * delta)


class is_on_cliff(Variable):
    value_type = bool
    entity = Person
    label = "is on a tax-benefit cliff"
    documentation = "Whether this person would be worse off if their employment income were higher."
    definition_period = YEAR

    def formula(person, period, parameters):
        return person("cliff_gap", period) > 0



from policyengine_us.model_api import *
from policyengine_core.periods import period as period_


def create_repeal_state_dependent_exemptions() -> Reform:
    class hi_regular_exemptions(Variable):
        value_type = float
        entity = TaxUnit
        label = "Hawaii regular exemptions"
        unit = USD
        documentation = (
            "https://files.hawaii.gov/tax/forms/2022/n11ins.pdf#page=20"
        )
        definition_period = YEAR
        defined_for = StateCode.HI

        def formula(tax_unit, period, parameters):
            exemptions_count = tax_unit("head_spouse_count", period)
            p = parameters(period).gov.states.hi.tax.income.exemptions
            # Aged heads and spouses get an extra base exemption.
            person = tax_unit.members
            head_or_spouse = person("is_tax_unit_head_or_spouse", period)
            aged = person("age", period) >= p.aged_threshold
            aged_head_spouse_count = tax_unit.sum(aged & head_or_spouse)
            total_exemption_count_including_aged = (
                exemptions_count + aged_head_spouse_count
            )
            return total_exemption_count_including_aged * p.base

    class md_total_personal_exemptions(Variable):
        value_type = float
        entity = TaxUnit
        label = "MD total personal exemptions"
        unit = USD
        definition_period = YEAR
        defined_for = StateCode.MD

        def formula(tax_unit, period, parameters):
            # Get md_personal_exemption from tax_unit multiplied by tax_unit_size
            md_personal_exemption = tax_unit("md_personal_exemption", period)
            tax_unit_size = tax_unit("head_spouse_count", period)
            return md_personal_exemption * tax_unit_size

    class mi_personal_exemptions(Variable):
        value_type = float
        entity = TaxUnit
        label = "Michigan personal and stillborn exemptions"
        defined_for = StateCode.MI
        unit = USD
        definition_period = YEAR
        reference = (
            "http://legislature.mi.gov/doc.aspx?mcl-206-30",
            "https://www.legislature.mi.gov/Publications/TaxpayerGuide.pdf",
        )

        def formula(tax_unit, period, parameters):
            p = parameters(period).gov.states.mi.tax.income.exemptions

            # Personal Exemptions & Stillborn Exemptions
            exemptions = add(
                tax_unit,
                period,
                ["head_spouse_count", "tax_unit_stillborn_children"],
            )
            return exemptions * p.personal

    class ne_exemptions(Variable):
        value_type = float
        entity = TaxUnit
        label = "Nebraska exemptions amount"
        unit = USD
        definition_period = YEAR
        reference = (
            "https://revenue.nebraska.gov/files/doc/tax-forms/2021/f_1040n_booklet.pdf"
            "https://revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_5.pdf"
        )
        defined_for = StateCode.NE

        def formula(tax_unit, period, parameters):
            p = parameters(period).gov.states.ne.tax.income.exemptions
            return tax_unit("head_spouse_count", period) * p.amount

    class oh_personal_exemptions_eligible_person(Variable):
        value_type = bool
        entity = Person
        label = "Eligible person for the Ohio Exemption Credit"
        definition_period = YEAR
        reference = (
            "https://tax.ohio.gov/static/forms/ohio_individual/individual/2021/pit-it1040-booklet.pdf#page=14",
        )
        defined_for = StateCode.OH

        def formula(person, period, parameters):
            # The personal exemption is provided for the head and spouse
            # if they are not claimed as a dependent elsewhere
            head_or_spouse = person("is_tax_unit_head_or_spouse", period)
            dependent_on_another_return = person(
                "claimed_as_dependent_on_another_return", period
            )
            # The personal exemption is also provided to dependents
            return ~dependent_on_another_return & head_or_spouse

    class ok_count_exemptions(Variable):
        value_type = float
        entity = TaxUnit
        label = "Count of Oklahoma exemptions"
        unit = USD
        definition_period = YEAR
        reference = (
            "https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/past-year/2021/511-Pkt-2021.pdf"
            "https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-Pkt.pdf"
        )
        defined_for = StateCode.OK

        def formula(tax_unit, period, parameters):
            p = parameters(period).gov.states.ok.tax.income.exemptions
            # special exemption AGI eligibility
            fagi = tax_unit("adjusted_gross_income", period)
            filing_status = tax_unit("filing_status", period)
            agi_eligible = fagi <= p.special_agi_limit[filing_status]
            # head exemptions
            age_eligible = (
                tax_unit("age_head", period) >= p.special_age_minimum
            )
            head_exemptions = where(
                tax_unit("blind_head", period), 2, 1
            ) + where(agi_eligible & age_eligible, 1, 0)
            # spouse exemptions
            age_eligible = (
                tax_unit("age_spouse", period) >= p.special_age_minimum
            )
            spouse_exemptions = where(
                filing_status == filing_status.possible_values.JOINT,
                (
                    where(tax_unit("blind_spouse", period), 2, 1)
                    + where(agi_eligible & age_eligible, 1, 0)
                ),
                0,
            )
            # dependent exemptions
            # total number of exemptions
            return head_exemptions + spouse_exemptions

    class ri_exemptions(Variable):
        value_type = float
        entity = TaxUnit
        label = "Rhode Island exemptions"
        unit = USD
        definition_period = YEAR
        reference = "https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022%20Tax%20Rate%20and%20Worksheets.pdf"
        defined_for = StateCode.RI

        def formula(tax_unit, period, parameters):
            p = parameters(period).gov.states.ri.tax.income.exemption

            exemptions_count = tax_unit("head_spouse_count", period)

            exemption_amount = exemptions_count * p.amount

            # Modified Federal AGI
            mod_agi = tax_unit("ri_agi", period)

            excess_agi = max_(0, mod_agi - p.reduction.start)

            increments = np.ceil(excess_agi / p.reduction.increment)

            reduction_rate = min_(p.reduction.rate * increments, 1)

            return exemption_amount * (1 - reduction_rate)

    class vt_personal_exemptions(Variable):
        value_type = float
        entity = TaxUnit
        label = "Vermont personal exemptions"
        unit = USD
        definition_period = YEAR
        defined_for = StateCode.VT
        reference = (
            "https://tax.vermont.gov/sites/tax/files/documents/IN-111-2022.pdf"
        )

        def formula(tax_unit, period, parameters):
            p = parameters(period).gov.states.vt.tax.income.exemption
            is_joint = tax_unit("tax_unit_is_joint", period)
            elsewhere_head = tax_unit("head_is_dependent_elsewhere", period)
            elsewhere_spouse = tax_unit(
                "spouse_is_dependent_elsewhere", period
            )
            eligible_head = (~elsewhere_head).astype(int)
            eligible_spouse = (~elsewhere_spouse).astype(int)
            eligible_count = eligible_head + (eligible_spouse * is_joint)
            # add number of other dependents claimed on federal Form 1040 (line 5c)
            total_exemption_count = eligible_count
            return total_exemption_count * p.personal

    class va_personal_exemption_person(Variable):
        value_type = float
        entity = Person
        label = "Virginia personal exemption for each person"
        defined_for = StateCode.VA
        unit = USD
        definition_period = YEAR
        reference = "https://law.lis.virginia.gov/vacodefull/title58.1/chapter3/article2/"

        def formula(person, period, parameters):
            head_or_spouse = person("is_tax_unit_head_or_spouse", period)
            amount = parameters(
                period
            ).gov.states.va.tax.income.exemptions.personal
            return amount * head_or_spouse

    class wv_personal_exemption(Variable):
        value_type = float
        entity = TaxUnit
        label = "West Virginia personal exemption"
        defined_for = StateCode.WV
        unit = USD
        definition_period = YEAR
        reference = "https://code.wvlegislature.gov/11-21/"

        def formula(tax_unit, period, parameters):
            p = parameters(period).gov.states.wv.tax.income.exemptions
            tax_unit_size = tax_unit("head_spouse_count", period)
            return where(
                tax_unit_size == 0, p.base_personal, p.personal * tax_unit_size
            )

    class ca_exemptions(Variable):
        value_type = float
        entity = TaxUnit
        label = "CA Exemptions"
        defined_for = StateCode.CA
        unit = USD
        definition_period = YEAR
        reference = "https://www.ftb.ca.gov/forms/2021/2021-540.pdf"

        def formula(tax_unit, period, parameters):
            p = parameters(period).gov.states.ca.tax.income.exemptions
            agi = tax_unit("adjusted_gross_income", period)
            filing_status = tax_unit("filing_status", period)
            # calculating phase out amount per credit
            over_agi_threshold = max_(
                0, agi - p.phase_out.start[filing_status]
            )
            increments = np.ceil(
                over_agi_threshold / p.phase_out.increment[filing_status]
            )
            exemption_reduction = increments * p.phase_out.amount
            # Personal Exemptions
            personal_exemption_count = p.personal_scale[filing_status]
            personal_aged_blind_exemption_count = (
                personal_exemption_count + tax_unit("aged_blind_count", period)
            )
            personal_aged_blind_exemption = max_(
                0,
                personal_aged_blind_exemption_count
                * (p.amount - exemption_reduction),
            )
            # Dependent exemptions
            # total exemptions
            return personal_aged_blind_exemption

    class ga_exemptions(Variable):
        value_type = float
        entity = TaxUnit
        label = "Georgia Exemptions"
        defined_for = StateCode.GA
        unit = USD
        definition_period = YEAR
        reference = (
            "https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500",
            "https://advance.lexis.com/documentpage/?pdmfid=1000516&crid=2c053fd5-32c1-4cc1-86b0-36aaade9da5b&pdistocdocslideraccess=true&config=00JAA1MDBlYzczZi1lYjFlLTQxMTgtYWE3OS02YTgyOGM2NWJlMDYKAFBvZENhdGFsb2feed0oM9qoQOMCSJFX5qkd&pddocfullpath=%2Fshared%2Fdocument%2Fstatutes-legislation%2Furn%3AcontentItem%3A6348-G0H1-DYB7-W3JT-00008-00&pdcomponentid=234187&pdtocnodeidentifier=ABWAALAADAAL&ecomp=k2vckkk&prid=4862391c-e031-443f-ad52-ae86c6bb5ce2",
        )

        def formula(tax_unit, period, parameters):
            p = parameters(period).gov.states.ga.tax.income.exemptions
            filing_status = tax_unit("filing_status", period)

            # Personal Exemptions
            personal_exemptions = p.personal[filing_status]

            # total exemptions
            return personal_exemptions

    class in_base_exemptions(Variable):
        value_type = float
        entity = TaxUnit
        label = "Indiana base exemptions"
        unit = USD
        definition_period = YEAR
        reference = "http://iga.in.gov/legislative/laws/2021/ic/titles/006#6-3-1-3.5"  # (a)(3)-(4)
        defined_for = StateCode.IN

        def formula(tax_unit, period, parameters):
            p = parameters(period).gov.states["in"].tax.income.exemptions
            return tax_unit("head_spouse_count", period) * p.base.amount

    class ia_exemption_credit(Variable):
        value_type = float
        entity = TaxUnit
        label = "Iowa exemption credit"
        unit = USD
        definition_period = YEAR
        reference = (
            "https://tax.iowa.gov/sites/default/files/2021-12/IA6251%2841131%29.pdf"
            "https://tax.iowa.gov/sites/default/files/2023-01/IA6251%2841131%29.pdf"
        )
        defined_for = StateCode.IA

        def formula(tax_unit, period, parameters):
            # count adult and dependent exemptions
            adult_count = tax_unit("head_spouse_count", period)
            filing_status = tax_unit("filing_status", period)
            hoh_status = filing_status.possible_values.HEAD_OF_HOUSEHOLD
            hoh_bonus = where(filing_status == hoh_status, 1, 0)
            # count extra adult exemptions based on being elderly and/or blind
            p = parameters(period).gov.states.ia.tax.income
            exemption = p.credits.exemption
            elder_head = tax_unit("age_head", period) >= exemption.elderly_age
            elder_spouse = (
                tax_unit("age_spouse", period) >= exemption.elderly_age
            )
            elder_count = elder_head.astype(int) + elder_spouse.astype(int)
            blind_head = tax_unit("blind_head", period)
            blind_spouse = tax_unit("blind_spouse", period)
            blind_count = blind_head.astype(int) + blind_spouse.astype(int)
            additional_count = elder_count + blind_count
            return (
                (adult_count + hoh_bonus) * exemption.personal
                + additional_count * exemption.additional
            )

    class ks_count_exemptions(Variable):
        value_type = float
        entity = TaxUnit
        label = "number of KS exemptions"
        unit = USD
        definition_period = YEAR
        reference = (
            "https://www.ksrevenue.gov/pdf/ip21.pdf"
            "https://www.ksrevenue.gov/pdf/ip22.pdf"
        )
        defined_for = StateCode.KS

        def formula(tax_unit, period, parameters):
            filing_status = tax_unit("filing_status", period)
            statuses = filing_status.possible_values
            joint = filing_status == statuses.JOINT
            hoh = filing_status == statuses.HEAD_OF_HOUSEHOLD
            adults = where(joint | hoh, 2, 1)
            return adults

    class ma_income_tax_exemption_threshold(Variable):
        value_type = float
        entity = TaxUnit
        label = "MA income tax exemption threshold"
        unit = USD
        documentation = "MA AGI threshold below which an individual is exempt from State income tax."
        definition_period = YEAR
        reference = "https://malegislature.gov/Laws/GeneralLaws/PartI/TitleIX/Chapter62/Section5"
        defined_for = StateCode.MA

        def formula(tax_unit, period, parameters):
            filing_status = tax_unit("filing_status", period)
            tax = parameters(period).gov.states.ma.tax.income
            exempt_status = tax.exempt_status.limit
            personal_exemptions_added = (
                exempt_status.personal_exemption_added[filing_status]
                * tax.exemptions.personal[filing_status]
            )
            return (
                exempt_status.base[filing_status] + personal_exemptions_added
            )

    # Using head and spouse count instead of exemptions count
    class wi_base_exemption(Variable):
        value_type = float
        entity = TaxUnit
        label = "Wisconsin base exemption"
        unit = USD
        definition_period = YEAR
        reference = (
            "https://www.revenue.wi.gov/TaxForms2021/2021-Form1f.pdf"
            "https://www.revenue.wi.gov/TaxForms2021/2021-Form1-Inst.pdf"
            "https://www.revenue.wi.gov/TaxForms2022/2022-Form1f.pdf"
            "https://www.revenue.wi.gov/TaxForms2022/2022-Form1-Inst.pdf"
            "https://docs.legis.wisconsin.gov/misc/lfb/informational_papers/january_2023/0002_individual_income_tax_informational_paper_2.pdf"
        )
        defined_for = StateCode.WI

        def formula(tax_unit, period, parameters):
            # compute base exemption amount
            p = parameters(period).gov.states.wi.tax.income
            return tax_unit("head_spouse_count", period) * p.exemption.base

    class de_personal_credit(Variable):
        value_type = float
        entity = TaxUnit
        label = "Delaware personal credit"
        unit = USD
        definition_period = YEAR
        reference = "https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf"
        defined_for = StateCode.DE

        def formula(tax_unit, period, parameters):
            p = parameters(period).gov.states.de.tax.income.credits
            head_spouse_count = tax_unit("head_spouse_count", period)
            return p.personal_credits.personal * head_spouse_count

    class ky_family_size_tax_credit_rate(Variable):
        value_type = float
        entity = TaxUnit
        label = "Kentucky family size tax credit rate"
        unit = "/1"
        definition_period = YEAR
        reference = "https://apps.legislature.ky.gov/law/statutes/statute.aspx?id=49188"
        defined_for = StateCode.KY

        def formula(tax_unit, period, parameters):
            income = tax_unit("ky_modified_agi", period)
            fpg = parameters(period).gov.hhs.fpg
            # This will be CONTIGUOUS_US for Kentucky.
            state_group = tax_unit.household("state_group_str", period)
            p1 = fpg.first_person[state_group]
            padd = fpg.additional_person[state_group]
            family_size = tax_unit("head_spouse_count", period)
            # No more than 4 people are accounted for in the credit
            p = parameters(period).gov.states.ky.tax.income.credits.family_size
            capped_family_size = min_(family_size, p.family_size_cap)
            poverty_index = p1 + padd * (capped_family_size - 1)
            share = income / poverty_index
            return p.rate.calc(share, right=True)

    class ok_child_care_child_tax_credit(Variable):
        value_type = float
        entity = TaxUnit
        label = "Oklahoma Child Care/Child Tax Credit"
        unit = USD
        definition_period = YEAR
        reference = (
            "https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/past-year/2021/511-Pkt-2021.pdf"
            "https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-Pkt.pdf"
        )
        defined_for = StateCode.OK

        def formula(tax_unit, period, parameters):
            p = parameters(period).gov.states.ok.tax.income.credits
            # determine AGI eligibility
            us_agi = tax_unit("adjusted_gross_income", period)
            agi_eligible = us_agi <= p.child.agi_limit
            # determine OK cdcc amount
            us_cdcc = tax_unit("cdcc", period)
            ok_cdcc = us_cdcc * p.child.cdcc_fraction
            # determine prorated fraction
            ok_agi = tax_unit("ok_agi", period)
            # Compute OK AGI as a share of US AGI.
            # Use a mask rather than where to avoid a divide-by-zero warning.
            agi_ratio = np.zeros_like(us_agi)
            mask = us_agi != 0
            agi_ratio[mask] = ok_agi[mask] / us_agi[mask]
            prorate = min_(1, max_(0, agi_ratio))
            # receive greater of OK cdcc or OK ctc amounts prorated if AGI eligible
            return agi_eligible * prorate * ok_cdcc

    class reform(Reform):
        def apply(self):
            self.neutralize_variable("al_dependent_exemption")
            self.neutralize_variable("il_dependent_exemption")
            self.neutralize_variable("la_dependents_exemption")
            self.neutralize_variable("mn_exemptions")
            self.neutralize_variable("ms_dependents_exemption")
            self.neutralize_variable("nj_dependents_exemption")
            self.neutralize_variable("ny_exemptions")
            self.neutralize_variable("sc_dependent_exemption")
            self.neutralize_variable("ut_personal_exemption")
            self.neutralize_variable("nc_child_deduction")
            self.neutralize_variable("nm_deduction_for_certain_dependents")
            self.neutralize_variable("mt_dependent_exemptions_person")
            self.neutralize_variable("az_dependent_tax_credit")
            self.neutralize_variable("ar_personal_credit_dependent")
            self.neutralize_variable("id_ctc")
            self.neutralize_variable("me_dependent_exemption_credit")
            self.update_variable(hi_regular_exemptions)
            self.update_variable(md_total_personal_exemptions)
            self.update_variable(mi_personal_exemptions)
            self.update_variable(ne_exemptions)
            self.update_variable(oh_personal_exemptions_eligible_person)
            self.update_variable(ok_count_exemptions)
            self.update_variable(ri_exemptions)
            self.update_variable(vt_personal_exemptions)
            self.update_variable(va_personal_exemption_person)
            self.update_variable(wv_personal_exemption)
            self.update_variable(ca_exemptions)
            self.update_variable(ga_exemptions)
            self.update_variable(in_base_exemptions)
            self.update_variable(ks_count_exemptions)
            self.update_variable(ma_income_tax_exemption_threshold)
            self.update_variable(wi_base_exemption)
            self.update_variable(ia_exemption_credit)
            self.update_variable(de_personal_credit)
            self.update_variable(ky_family_size_tax_credit_rate)
            self.update_variable(ok_child_care_child_tax_credit)

    return reform


def create_repeal_state_dependent_exemptions_reform(
    parameters, period, bypass: bool = False
):
    if bypass:
        return create_repeal_state_dependent_exemptions()

    p = parameters.gov.contrib.repeal_state_dependent_exemptions

    reform_active = False
    current_period = period_(period)

    for i in range(5):
        if p(current_period).in_effect:
            reform_active = True
            break
        current_period = current_period.offset(1, "year")

    if reform_active:
        return create_repeal_state_dependent_exemptions()
    else:
        return None


repeal_state_dependent_exemptions = (
    create_repeal_state_dependent_exemptions_reform(None, None, bypass=True)
)



from policyengine_us.model_api import *


class qualified_reit_and_ptp_income(Variable):
    value_type = float
    entity = Person
    label = "REIT and PTP Income"
    unit = USD
    documentation = "REIT and Publicly Traded Partnership Income. Part of the QBID calculation."
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/1#h_11"
    uprating = "calibration.gov.irs.soi.qualified_dividend_income"



from policyengine_us.model_api import *


class taxsim_ui(Variable):
    value_type = float
    entity = TaxUnit
    label = "Unemployment compensation (spouse)"
    unit = USD
    documentation = "Unemployment compensation received - secondary taxpayer. The split is relevant only 2020-2021."
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        person = tax_unit.members
        is_spouse = person("is_tax_unit_spouse", period)
        unemployment_compensation = person("unemployment_compensation", period)
        return tax_unit.sum(unemployment_compensation * is_spouse)



from policyengine_us.model_api import *


class oh_pension_based_retirement_income_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "Ohio pension based retirement income credit"
    unit = USD
    definition_period = YEAR
    reference = "https://codes.ohio.gov/ohio-revised-code/section-5747.055"
    defined_for = "oh_pension_based_retirement_income_credit_eligible"

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.oh.tax.income.credits.retirement.pension_based

        person = tax_unit.members
        pension_income = person("taxable_pension_income", period)
        head_or_spouse = person("is_tax_unit_head_or_spouse", period)
        eligible_pension = pension_income * head_or_spouse
        total_pension_income = tax_unit.sum(eligible_pension)

        return p.amount.calc(total_pension_income, right=True)



from policyengine_us.model_api import *


class nm_eitc_demographic_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Meets demographic eligibility for New Mexico EITC"
    definition_period = YEAR
    defined_for = StateCode.NM

    def formula(tax_unit, period, parameters):
        # New Mexico applies the same criteria as the federal EITC, but
        # changes the minimum age.
        person = tax_unit.members
        has_child = tax_unit("tax_unit_children", period) > 0
        age = person("age", period)
        # Relative parameter reference break branching in some states that
        # modify EITC age limits.
        min_age = parameters.gov.states.nm.tax.income.credits.eitc.eligibility.age.min(
            period
        )
        max_age = parameters.gov.irs.credits.eitc.eligibility.age.max(period)
        meets_age_requirements = (age >= min_age) & (age <= max_age)
        return has_child | tax_unit.any(meets_age_requirements)



from policyengine_us.model_api import *


class meets_snap_work_exception(Variable):
    value_type = bool
    entity = Person
    label = "Meets SNAP student work exception"
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/7/2015#e_4"

    def formula(person, period, parameters):
        # Exception 4: Employed at least 20 hours per week or work-study
        hours_worked = person("weekly_hours_worked_before_lsr", period)
        p = parameters(period).gov.usda.snap.student
        meets_hours_requirement = hours_worked >= p.working_hours_threshold

        # Or participating in federal/state work-study
        participates_in_work_study = person(
            "is_federal_work_study_participant", period
        )

        return meets_hours_requirement | participates_in_work_study



from policyengine_us.model_api import *
from policyengine_core.periods import period as period_


def create_nyc_mamdani_income_tax() -> Reform:
    class nyc_mamdani_income_tax(Variable):
        value_type = float
        entity = Person
        label = "Zohran Mamdani NYC income tax"
        unit = USD
        definition_period = YEAR
        defined_for = "in_nyc"

        def formula(person, period, parameters):
            taxable_income = person.tax_unit("nyc_taxable_income", period)
            p = parameters(period).gov.local.ny.mamdani_income_tax
            return p.rate.calc(taxable_income)

    class nyc_income_tax_before_credits(Variable):
        value_type = float
        entity = TaxUnit
        label = "NYC income tax before credits"
        unit = USD
        definition_period = YEAR
        defined_for = "in_nyc"

        def formula(tax_unit, period, parameters):
            taxable_income = tax_unit("nyc_taxable_income", period)
            filing_status = tax_unit("filing_status", period)
            filing_statuses = filing_status.possible_values
            p = parameters(period).gov.local.ny.nyc.tax.income.rates
            regular_tax = select(
                [
                    filing_status == filing_statuses.SINGLE,
                    filing_status == filing_statuses.JOINT,
                    filing_status == filing_statuses.HEAD_OF_HOUSEHOLD,
                    filing_status == filing_statuses.SURVIVING_SPOUSE,
                    filing_status == filing_statuses.SEPARATE,
                ],
                [
                    p.single.calc(taxable_income),
                    p.joint.calc(taxable_income),
                    p.head_of_household.calc(taxable_income),
                    p.surviving_spouse.calc(taxable_income),
                    p.separate.calc(taxable_income),
                ],
            )
            mamdani_tax = add(tax_unit, period, ["nyc_mamdani_income_tax"])
            return regular_tax + mamdani_tax

    class reform(Reform):
        def apply(self):
            self.update_variable(nyc_mamdani_income_tax)
            self.update_variable(nyc_income_tax_before_credits)

    return reform


def create_nyc_mamdani_income_tax_reform(
    parameters, period, bypass: bool = False
):
    if bypass:
        return create_nyc_mamdani_income_tax()

    p = parameters.gov.local.ny.mamdani_income_tax

    reform_active = False
    current_period = period_(period)

    for i in range(5):
        if p(current_period).in_effect:
            reform_active = True
            break
        current_period = current_period.offset(1, "year")

    if reform_active:
        return create_nyc_mamdani_income_tax()
    else:
        return None


nyc_mamdani_income_tax = create_nyc_mamdani_income_tax_reform(
    None, None, bypass=True
)



from policyengine_us.model_api import *


class va_reduced_itemized_deductions(Variable):
    value_type = float
    entity = TaxUnit
    label = "Virginia reduced itemized deductions"
    unit = USD
    definition_period = YEAR
    reference = "https://www.tax.virginia.gov/sites/default/files/taxforms/individual-income-tax/2021/schedule-2021.pdf"
    defined_for = StateCode.VA

    def formula(tax_unit, period, parameters):
        p_va = parameters(period).gov.states.va.tax.income.deductions.itemized

        # Part A: If AGI from federal return is over a certain amount, then
        # limitations to the itemized deduction are applied
        # Line 1 - sum of medical exepens ded., capped state and local tax,
        # interest ded., charitable ded., and casualty loss ded.
        applicable_ded = add(
            tax_unit,
            period,
            p_va.reduction.applicable,
        )
        # Line 2 medical expense ded., interest ded., casualty loss ded., and gambling losses
        reducible_ded = add(
            tax_unit,
            period,
            p_va.reduction.reducible,
        )
        # Line 3 - subtract Line 2 from line 1
        excess_ded = max_(applicable_ded - reducible_ded, 0)
        # If 0 - no reduction
        # Line 4 - IRS deduction rate of excess

        # The federal limitation on itemized deductions does not apply during the TCJA period
        # Virginia still applies the limitation

        # Here, we use a parameter that mostly keeps in line with the IRS's;
        # Virginia follows the IRS, except for TCJA-related policies, hence we
        # remove anything TCJA related, but keep everything else coupled to the IRS

        # Shortcut to the full parameter, not parameter at instant, for IRS itemized limitations
        full_irs_param: Parameter = (
            parameters.gov.irs.deductions.itemized.limitation
        )

        va_itemized_deduction_rate = (
            full_irs_param.itemized_deduction_rate.clone()
        )

        # Eliminate infinite values created by the TCJA's elimination of this param
        va_itemized_deduction_rate.values_list = [
            entry
            for entry in va_itemized_deduction_rate.values_list
            if not (entry.value == np.inf or entry.value == -np.inf)
        ]

        excess_ded_fraction = excess_ded * va_itemized_deduction_rate(period)
        # Line 5 Federal AGI
        federal_agi = tax_unit("adjusted_gross_income", period)
        # Line 6 - applicable amount
        filing_status = tax_unit("filing_status", period)
        applicable_amount = p_va.applicable_amount[filing_status]
        # Line 7 - excess
        excess = max_(federal_agi - applicable_amount, 0)
        # Line 8 - excess times federal item. ded. AGI rate

        # Here again, Virginia follows the IRS, except for TCJA-related policies,
        # and we'll need to eliminate infinite values created by TCJA abolishing param
        va_agi_rate = full_irs_param.agi_rate.clone()
        va_agi_rate.values_list = [
            entry
            for entry in va_agi_rate.values_list
            if not (entry.value == np.inf or entry.value == -np.inf)
        ]

        agi_adjustment = va_agi_rate(period) * excess
        # Line 9 - min of line 4 and line 8
        va_itm_deds_adjustment = min_(agi_adjustment, excess_ded_fraction)
        # Output from Part A (Line 1 - Line 9)
        # Line 17 of the VA Schedule A
        reduced_selected_ded = max_(applicable_ded - va_itm_deds_adjustment, 0)

        # Part B: state and local income tax modification
        # Form Part A - line 11 - Line 9 divided by line 3
        adjustment_fraction = np.zeros_like(excess_ded)
        mask = excess_ded != 0
        adjustment_fraction[mask] = (
            va_itm_deds_adjustment[mask] / excess_ded[mask]
        )
        # Part B Line 13 - capped state and local income tax

        # Line 14 - Multiply line 11 by line 13
        capped_state_and_local_tax = tax_unit(
            "va_capped_state_and_local_sales_or_income_tax", period
        )

        state_and_local_tax_adj = (
            capped_state_and_local_tax * adjustment_fraction
        )
        # Line 15 - Subtract Line 14 from Line 13
        reduced_state_and_local_tax = max_(
            capped_state_and_local_tax - state_and_local_tax_adj, 0
        )
        return max_(reduced_selected_ded - reduced_state_and_local_tax, 0)



from policyengine_us.model_api import *


class purchased_qualifying_used_clean_vehicle(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Purchased a qualifying used clean vehicle"
    definition_period = YEAR



from policyengine_us.model_api import *


def fpg(unit_size, state_group, period, parameters):
    p_fpg = parameters(period).gov.hhs.fpg
    p1 = p_fpg.first_person[state_group]
    pn = p_fpg.additional_person[state_group]
    return p1 + pn * (unit_size - 1)


class tax_unit_fpg(Variable):
    value_type = float
    entity = TaxUnit
    label = "Tax unit's federal poverty guideline"
    definition_period = YEAR
    unit = USD

    def formula(tax_unit, period, parameters):
        n = tax_unit("tax_unit_size", period)
        state_group = tax_unit.household("state_group_str", period)
        return fpg(n, state_group, period, parameters)



from policyengine_us.model_api import *


class mt_aged_exemption_eligible_person(Variable):
    value_type = bool
    entity = Person
    label = "Montana aged exemptions when married couples file separately"
    definition_period = YEAR
    reference = "https://regulations.justia.com/states/montana/department-42/chapter-42-15/subchapter-42-15-4/rule-42-15-402/"
    defined_for = StateCode.MT

    def formula(person, period, parameters):
        p = parameters(period).gov.states.mt.tax.income.exemptions

        if p.applies:
            age = person("age", period)
            aged = age >= p.age_threshold
            return aged & person("is_tax_unit_head_or_spouse", period)

        return False



from policyengine_us.model_api import *


class taxsim_pbusinc(Variable):
    value_type = float
    entity = TaxUnit
    label = "QBI for the primary taxpayer (TAXSIM)"
    unit = USD
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        person = tax_unit.members
        qbi = person("qualified_business_income", period)
        is_head = person("is_tax_unit_head", period)
        return tax_unit.sum(qbi * is_head)



from policyengine_us.model_api import *


class disability_benefits(Variable):
    value_type = float
    entity = Person
    label = "disability benefits"
    unit = USD
    documentation = "Disability benefits from employment (not Social Security), except for worker's compensation."
    definition_period = YEAR



from policyengine_us.model_api import *


class nh_base_exemption(Variable):
    value_type = float
    entity = TaxUnit
    label = "New Hampshire base exemption"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.NH

    def formula(tax_unit, period, parameters):
        filing_status = tax_unit("filing_status", period)
        p = parameters(period).gov.states.nh.tax.income.exemptions.amount
        return p.base[filing_status]



from policyengine_us.model_api import *


class snap_take_up_seed(Variable):
    value_type = float
    entity = SPMUnit
    label = "Randomly assigned seed for SNAP take-up"
    definition_period = YEAR



from policyengine_us.model_api import *


class il_bcc_eligible(Variable):
    value_type = bool
    entity = Person
    label = "Illinois Health Benefits for Persons with Breast or Cervical Cancer eligible"
    definition_period = YEAR
    reference = (
        "https://www.dhs.state.il.us/page.aspx?item=33528",
        "https://www.dhs.state.il.us/page.aspx?item=33527",
    )
    defined_for = StateCode.IL

    def formula(person, period, parameters):
        is_female = person("is_female", period)
        age_eligible = person("il_bcc_age_eligible", period)
        immigration_eligible = person(
            "il_hfs_immigration_status_eligible", period
        )
        insurance_eligible = person("il_bcc_insurance_eligible", period)
        return (
            is_female
            & age_eligible
            & immigration_eligible
            & insurance_eligible
        )



from policyengine_us.model_api import *
from policyengine_core.periods import period as period_
from policyengine_core.periods import instant


def create_ctc_per_child_phase_in() -> Reform:
    class ctc_phase_in_relevant_earnings(Variable):
        value_type = float
        entity = TaxUnit
        label = "Relevant earnings for the CTC phase-in"
        unit = USD
        definition_period = YEAR
        reference = "https://www.law.cornell.edu/uscode/text/26/24#d"

        def formula(tax_unit, period, parameters):

            ctc = parameters(period).gov.irs.credits.ctc

            earnings = tax_unit("tax_unit_earned_income", period)
            earnings_over_threshold = max_(
                0, earnings - ctc.refundable.phase_in.threshold
            )
            qualifying_children = tax_unit("ctc_qualifying_children", period)
            phase_in_rate = ctc.refundable.phase_in.rate * qualifying_children
            return earnings_over_threshold * phase_in_rate

    class reform(Reform):
        def apply(self):
            self.update_variable(ctc_phase_in_relevant_earnings)

    return reform


def create_ctc_per_child_phase_in_reform(
    parameters, period, bypass: bool = False
):
    if bypass:
        return create_ctc_per_child_phase_in()

    p = parameters.gov.contrib.ctc.per_child_phase_in

    reform_active = False
    current_period = period_(period)
    for i in range(5):
        if p(current_period).in_effect:
            reform_active = True
            break
        current_period = current_period.offset(1, "year")

    if reform_active:
        return create_ctc_per_child_phase_in()
    else:
        return None


ctc_per_child_phase_in = create_ctc_per_child_phase_in_reform(
    None, None, bypass=True
)



from policyengine_us.model_api import *


class school_meal_paid_daily_subsidy(Variable):
    value_type = float
    entity = SPMUnit
    definition_period = YEAR
    label = "School meal subsidies per full-price child per day"
    unit = USD
    documentation = "Value of school meal subsidies paid to full-price children per day in household's state"

    def formula(spm_unit, period, parameters):
        # Get state group and tier (based on poverty ratio) for SPM unit.
        state_group = spm_unit.household("state_group_str", period)
        p_amount = parameters(period).gov.usda.school_meals.amount
        nslp_per_child = p_amount.nslp[state_group]["PAID"]
        sbp_per_child = p_amount.sbp[state_group]["PAID"]
        return nslp_per_child + sbp_per_child



from policyengine_us.model_api import *


class oh_lump_sum_distribution_credit_eligible_person(Variable):
    value_type = bool
    entity = Person
    label = "Eligible person for the Ohio lump sum distribution credit"
    definition_period = YEAR
    reference = (
        "https://tax.ohio.gov/static/forms/ohio_individual/individual/2021/pit-it1040-booklet.pdf#page=29",
    )
    defined_for = StateCode.OH

    def formula(person, period, parameters):
        p = parameters(
            period
        ).gov.states.oh.tax.income.credits.lump_sum_distribution
        is_head_or_spouse = person("is_tax_unit_head_or_spouse", period)
        age_eligible = person("age", period) >= p.age_threshold
        return is_head_or_spouse & age_eligible



from policyengine_us.model_api import *


class medicare_quarters_of_coverage(Variable):
    value_type = int
    entity = Person
    label = "Medicare quarters of coverage"
    documentation = (
        "Number of quarters of Medicare-covered employment. Most people have "
        "40+ quarters and qualify for premium-free Part A. Those with 30-39 "
        "quarters pay a reduced premium, and those with fewer than 30 quarters "
        "pay the full premium."
    )
    definition_period = YEAR

    adds = "gov.hhs.medicare.part_a.default_quarters_of_coverage"



from policyengine_us.model_api import *


class ma_ccfa_total_copay(Variable):
    value_type = float
    entity = SPMUnit
    unit = USD
    label = "Massachusetts Child Care Financial Assistance (CCFA) parent total copay"
    definition_period = MONTH
    defined_for = StateCode.MA
    reference = "https://www.mass.gov/doc/eecs-financial-assistance-policy-guide-february-1-2022/download#page=76"

    def formula(spm_unit, period, parameters):
        # TAFDC recipients get free child care (no copay)
        tafdc_eligible = spm_unit("ma_tafdc_eligible", period)

        p = parameters(period).gov.states.ma.eec.ccfa.copay.ratio
        base_copay = spm_unit("ma_ccfa_base_copay", period)
        eligible_child = spm_unit.members("ma_ccfa_eligible_child", period)
        num_children = spm_unit.sum(eligible_child)

        first_child_fee = (num_children >= 1) * (base_copay * p.first_child)
        second_child_fee = (num_children >= 2) * (base_copay * p.second_child)

        additional_children_count = max_(0, num_children - 2)
        additional_children_fee = additional_children_count * (
            base_copay * p.additional_child
        )

        total_copay = (
            first_child_fee + second_child_fee + additional_children_fee
        )

        # Return zero copay for TAFDC recipients
        return where(tafdc_eligible, 0, total_copay)



from policyengine_us.model_api import *


class tenant_pays_utilities(Variable):
    value_type = bool
    entity = Household
    label = "Whether the tenant is responsible for utility payments"
    definition_period = YEAR
